/*
 * This file is part of faultd.
 *
 * Copyright © 2017 Samsung Electronics
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <sys/stat.h>

#include "service.h"
#include "json-config.h"
#include "log.h"

static const char *dest_path = "/tmp/";

static int create_dir(const char *path)
{
	if (mkdir(path, 0755) != 0) {
		if (errno == EEXIST)
			return 0;
		else
			return -errno;
	}

	return 0;
}

static int create_vip_service_override(const char *service_name)
{
	FILE *fp;
	char *location;
	char *service_override;
	int ret;

	ret = asprintf(&location, "%s/%s.d", dest_path, service_name);
	if (ret == -1)
		return -errno;

	ret = create_dir(location);
	if (ret != 0) {
		free(location);
		return ret;
	}

	ret = asprintf(&service_override, "%s/faultd.conf", location);
	free(location);
	if (ret == -1)
		return -errno;

	fp = fopen(service_override, "w");
	free(service_override);
	if (fp == NULL)
		return -errno;

	fprintf(fp, "[Service]\n"
				"Restart=no\n");
	fclose(fp);

	return 0;
}

int main(int argc, char *argv[])
{
	struct systemd_service s;
	struct json_object *services_config = json_object_new_object();
	int ret;

	/*
	 * This program follows systemd generator specification, please see:
	 * https://www.freedesktop.org/software/systemd/man/systemd.generator.html
	 */

	if (argc > 1 && argc != 4) {
		log_error("This program takes three or no arguments");
		return EXIT_FAILURE;
	}

	if (argc > 1)
		dest_path = argv[2];

	/* load services configuration into json object */
	ret = load_config(FAULTD_DEFAULT_SERVICES_CONFIG_PATH, CONFIG_TYPE_DIR, 0,
					  services_config);
	if (ret) {
		log_error("Unable to load services config %s", FAULTD_DEFAULT_SERVICES_CONFIG_PATH);
		json_object_put(services_config);
		return ret;
	}

	ret = load_config(FAULTD_CUSTOM_SERVICES_CONFIG_PATH, CONFIG_TYPE_DIR, 1,
					  services_config);
	if (ret) {
		log_error("Unable to load services config %s", FAULTD_CUSTOM_SERVICES_CONFIG_PATH);
		json_object_put(services_config);
		return ret;
	}

	/*
	 * iterate through every config node and create override unit file if
	 * the service is of VIP type
	 */

	json_object_object_foreach(services_config, service_name, node) {
		s.service_type = NULL;

		ret = get_config_field(node, "ServiceType", &s.service_type, json_type_string);
		if (ret) {
			log_error("Unable to get service type for %s", service_name);
			continue;
		}

		if (systemd_service_is_of_type(&s, FAULTD_SERVICE_TYPE_VIP)) {
			ret = create_vip_service_override(service_name);
			if (ret)
				log_error("Unable to create override for VIP service %s: %s\n",
						  service_name, strerror(ret));
		}

		free(s.service_type);
	}

	json_object_put(services_config);
	return 0;
}
