/*
 * This file is part of faultd.
 *
 * Copyright © 2017 Samsung Electronics
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <errno.h>

#include "action_executor.h"
#include "decision_made_event.h"
#include "log.h"

struct action_executor {
	struct faultd_event_handler handler;
	struct list_head builtin_actions;
	sd_event *loop;
};

static int ae_event_match(struct faultd_event_handler *handler,
						  struct faultd_event *ev)
{
	/* We are interested only in decision made events */
	return faultd_event_is_of_type(ev, DECISION_MADE_EVENT_ID);
}

static int ae_execute_action(struct faultd_event_handler *handler)
{
	struct faultd_event *ev = pop_faultd_event(&handler->event_queue);
	struct decision_made_event *dm_ev = to_decision_made_event(ev);
	struct action_executor *ae = container_of(handler, struct action_executor,
											  handler);
	struct faultd_action *action;
	int ret = -1;

	list_for_each_entry(action, &ae->builtin_actions, node)
		if (strcmp(action->action_id , dm_ev->action) == 0) {
			ret = append_faultd_event(&action->execution_queue, ev);
			if (ret) {
				log_error("Unable to append event to the queue");
				ret = 0;
				goto cleanup;
			}
			break;
		}

	/* TODO: Here is a good place to insert plugin search */

	if (ret) {
		log_error("No implementation of %s action.", dm_ev->action);
		ret = 0;
		goto cleanup;
	}

cleanup:
	faultd_event_unref(ev);
	return ret;
}

static struct action_executor action_executor = {
	.handler = {
		.name = "action_executor_event_handler",
		.event_match = ae_event_match,
		.handle_event = ae_execute_action,

		.node = LIST_HEAD_INIT(action_executor.handler.node),
	},
	.builtin_actions = LIST_HEAD_INIT(action_executor.builtin_actions),
};

FAULTD_EVENT_HANDLER_REGISTER(action_executor.handler,
							  action_executor_eh,
							  FAULTD_MODULE_TYPE_ACTION_EXECUTOR)

static int action_executor_callback(sd_event_source *s, int fd,
									uint32_t revents, void *userdata)
{
	struct faultd_action *act = userdata;
	struct ae_event_data ae_data = {
		.reason = NULL,
		.action = act->action_id,
		.action_impl = act->impl_name,
		.result = -ENODEV,
	};
	struct faultd_event *ev = NULL;
	struct action_executed_event *exe_info;
	int action_ret;
	int ret;

	ret = faultd_object_new(&ae_data.action_log);
	if (ret < 0) {
		log_error("Unable to init faultd object");
		ret = -ENOMEM;
		goto out;
	}

	ret = faultd_event_create(ACTION_EXECUTED_EVENT_ID, &ae_data, &ev);
	if (ret < 0) {
		log_error("Unable to create action executed event");
		faultd_object_destroy(ae_data.action_log);
		goto out;
	}

	exe_info = to_action_executed_event(ev);

	/*
	 * If action is not fully executed after returning from this function
	 * it takes the ownership of of exe_info and should report it when
	 * action execution is completed.
	 */
	action_ret = act->execute(act, exe_info);
	if (exe_info->result != -EPROBE_DEFER) {
		if (exe_info->result < 0)
			log_error_errno(exe_info->result, "Unable to execute action %s",
							ae_data.action);

		ret = event_processor_report_event(ev);
		if (ret) {
			log_error_errno(ret, "Unable to report AE event");
			ret = 0;
			goto cleanup;
		}
	}

	ret = action_ret;

cleanup:
	faultd_object_unref(ae_data.action_log);
	faultd_event_unref(ev);
out:
	return ret;
}

int action_executor_action_register(struct faultd_action *act)
{
	int ret;

	if (!act->action_id || !act->execute)
		return -EINVAL;

	ret = init_notify_queue_head(&act->execution_queue);
	if (ret)
		return ret;

	ret = sd_event_add_io(sd_event_source_get_event(
							  action_executor.handler.event_queue_sd_source),
						  &act->execution_queue_sd_source,
						  act->execution_queue.fd, EPOLLIN,
						  action_executor_callback, act);
	if (ret < 0) {
		log_error("Unable to add event io %d", ret);
		goto cleanup_queue;
	}

	list_add_tail(&act->node, &action_executor.builtin_actions);
	return 0;

cleanup_queue:
	cleanup_notify_queue_head(&act->execution_queue);

	return ret;
}

void action_executor_action_unregister(struct faultd_action *act)
{
	int ret;

	/* We log errors but go forward to cleanup everything we can */
	ret = sd_event_source_set_enabled(act->execution_queue_sd_source,
									  SD_EVENT_OFF);
	if (ret < 0)
		log_error("Could not disable queue event source %d", ret);

	sd_event_source_unref(act->execution_queue_sd_source);

	/* TODO: What to do with all pending executions? */
	while (!nqueue_empty(&act->execution_queue)) {
		struct faultd_event *ev = pop_faultd_event(&act->execution_queue);
		char *str = faultd_event_to_string(ev);

		log_error("Unhandled event:\n%s", str);
		free(str);
		faultd_event_unref(ev);
	}

	list_del(&act->node);

}
