/*
 * This file is part of faultd.
 *
 * Copyright © 2017 Samsung Electronics
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <systemd/sd-bus.h>
#include <systemd/sd-event.h>

#include "action.h"
#include "action_executor.h"
#include "decision_made_event.h"
#include "log.h"
#include "systemd_dbus.h"
#include "common.h"

struct priv_data {
	struct action_executed_event *exe_info;
	char *job;
	char *service_path;
	char *recovery_unit;
	sd_bus_slot *slot;
};

static int simple_restart_service(char *service_path,
								  struct action_executed_event *exe_info)
{
	int ret;

	ret = faultd_dbus_call_systemd_simple(service_path,
										  SYSTEMD_UNIT_INTERFACE,
										  "Restart",
										  "s",
										  "replace");
	if (ret < 0)
		faultd_object_append_string(exe_info->action_log, "error",
									"Failed to call systemd");
	else
		log_kmsg("Restarted service: %s %d", service_path, ret);

	exe_info->result = ret;

	return 0;
}

static int on_job_removed(sd_bus_message *m, void *userdata,
						  sd_bus_error *ret_error)
{
	struct priv_data *priv_data = userdata;
	uint32_t uid;
	char *job;
	char *unit;
	char *result;
	int ret = 0;

	if (strcmp(SYSTEMD_MANAGER_INTERFACE, sd_bus_message_get_interface(m)))
		goto finish;

	if (strcmp("JobRemoved", sd_bus_message_get_member(m)))
		goto finish;

	ret = sd_bus_message_read(m, "uoss", &uid, &job, &unit, &result);
	if (ret < 0) {
		log_error("Malformed systemd signal");
		return -EINVAL;
	}

	if (strcmp(priv_data->job, job) == 0) {
		ret = simple_restart_service(priv_data->service_path, priv_data->exe_info);

		/* TODO: improve error handling */
		event_processor_report_event(&priv_data->exe_info->event);
		faultd_event_unref(&priv_data->exe_info->event);
		free(priv_data->job);
		sd_bus_slot_unref(priv_data->slot);

		log_kmsg("Recovered service: %s", priv_data->service_path);
		free(priv_data);
	} else {
		ret = 0;
	}
finish:
	return ret;
}

static int recover_service(struct faultd_action *action,
						   struct action_executed_event *exe_info)
{
	struct faultd_event *ev = pop_faultd_event(&action->execution_queue);
	struct decision_made_event *dm_ev = to_decision_made_event(ev);
	struct priv_data *priv_data;
	_cleanup_(sd_bus_unrefp) sd_bus *bus = NULL;
	_cleanup_(sd_bus_message_unrefp) sd_bus_message *reply = NULL;
	_cleanup_(sd_bus_error_free) sd_bus_error error = SD_BUS_ERROR_NULL;
	const char *addr;
	const char *job;
	char *service_path = NULL;
	char *recovery_unit = NULL;
	int ret;

	/*
	 * We are passing an event to exe_info, so there is no need to unref it
	 * in this function
	 */
	exe_info->reason = ev;

	ret = faultd_object_get_string(dm_ev->action_data, FAULTD_AD_SERVICE_NAME,
								   &service_path);
	if (ret < 0) {
		faultd_object_append_string(exe_info->action_log, "error",
									"Service path has not been provided");
		exe_info->result = -EINVAL;
		return 0;
	}

	ret = faultd_object_get_string(dm_ev->action_data, FAULTD_AD_CLEANUP_UNIT,
								   &recovery_unit);
	if (ret < 0) {
		faultd_object_append_string(exe_info->action_log, "warning",
									"Cleanup service has not been provided");
		goto simple_reset;
	}

	priv_data = calloc(1, sizeof(*priv_data));
	if (!priv_data) {
		faultd_object_append_string(exe_info->action_log, "error",
						   "Unable to allocate memory. Trying simple reset.");
		goto simple_reset;
	}

	priv_data->exe_info = exe_info;
	priv_data->recovery_unit = recovery_unit;
	priv_data->service_path = service_path;

	ret = faultd_acquire_systemd_bus(&bus);
	if (ret < 0) {
		faultd_object_append_string(exe_info->action_log, "error",
									"Unable to acquire systemd bus");
		exe_info->result = ret;
		goto free_priv_data;
	}

	ret = sd_bus_get_address(bus, &addr);
	if (ret < 0) {
		faultd_object_append_string(exe_info->action_log, "error",
									"Unable to get bus address");
		exe_info->result = ret;
		goto free_priv_data;
	}

	if (strcmp(SYSTEMD_PRIVATE_BUS, addr) == 0)
		ret = sd_bus_add_filter(bus,
							   &priv_data->slot,
							   on_job_removed,
							   priv_data);
	else
		ret = sd_bus_add_match(bus,
							  &priv_data->slot,
							  "type='signal',sender='org.freedesktop.systemd1',"
							  "interface='org.freedesktop.systemd1.Manager',"
							  "member='JobRemoved',"
							  "path_namespace='/org/freedesktop/systemd1'",
							  on_job_removed,
							  priv_data);
	if (ret < 0) {
		faultd_object_append_string(exe_info->action_log, "error",
									"Unable to register for JobRemoved signal");
		exe_info->result = ret;
		goto free_priv_data;
	}

	ret = sd_bus_call_method(bus,
							 SYSTEMD_SERVICE,
							 SYSTEMD_OBJ,
							 SYSTEMD_MANAGER_INTERFACE,
							 "StartUnit",
							 &error,
							 &reply,
							 "ss",
							 recovery_unit,
							 "replace");
	if (ret < 0) {
		faultd_object_append_string(exe_info->action_log, "error",
									"Unable to register for JobRemoved signal");
		exe_info->result = ret;
		goto unregister_handler;
	}

	ret = sd_bus_message_read(reply, "o", &job);
	if (ret < 0) {
		faultd_object_append_string(exe_info->action_log, "warning",
									"Malformed systemd reply. Unable to identify job.");
		exe_info->result = ret;
		goto unregister_handler;
	}

	priv_data->job = strdup(job);
	if (!priv_data->job) {
		faultd_object_append_string(exe_info->action_log, "error",
									"Unable to duplicate job id");
		exe_info->result = ret;
		goto unregister_handler;
	}

	exe_info->result = -EPROBE_DEFER;
	faultd_event_ref(&exe_info->event);
	return 0;

unregister_handler:
	sd_bus_slot_unref(priv_data->slot);

free_priv_data:
	free(priv_data);

	return 0;

simple_reset:
	return simple_restart_service(service_path, exe_info);
}

static struct faultd_action service_recover_action = {
	.action_id = FAULTD_ACTION_SERVICE_RECOVER_ID,
	.impl_name = FAULTD_DEFAULT_ACTION_IMPL,
	.execute = recover_service,
	.node = LIST_HEAD_INIT(service_recover_action.node),
};

FAULTD_ACTION_REGISTER_SIMPLE(service_recover_action);
