/*
 * This file is part of faultd.
 *
 * Copyright © 2017 Samsung Electronics
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "action.h"
#include "action_executor.h"
#include "decision_made_event.h"
#include "log.h"
#include "systemd_dbus.h"
#include "common.h"

static int restart_service(struct faultd_action *action,
						   struct action_executed_event *exe_info)
{
	struct faultd_event *ev = pop_faultd_event(&action->execution_queue);
	struct decision_made_event *dm_ev = to_decision_made_event(ev);
	char *service_path = NULL;
	int ret;

	/*
	 * We are passing an event to exe_info, so there is no need to unref it
	 * in this function
	 */
	exe_info->reason = ev;

	ret = faultd_object_get_string(dm_ev->action_data,
								   FAULTD_AD_SERVICE_NAME, &service_path);
	if (!service_path) {
		faultd_object_append_string(exe_info->action_log, "error",
									"Service path has not been provided");
		exe_info->result = -EINVAL;
		return 0;
	}

	ret = faultd_dbus_call_systemd_simple(service_path,
										  SYSTEMD_UNIT_INTERFACE,
										  "Restart",
										  "s",
										  "replace");
	if (ret < 0)
		faultd_object_append_string(exe_info->action_log, "error",
									"Failed to call systemd");
	else
		log_kmsg("Restarting service: %s", service_path);

	exe_info->result = ret;

	return 0;
}

static struct faultd_action service_restart_action = {
	.action_id = FAULTD_ACTION_SERVICE_RESTART_ID,
	.impl_name = FAULTD_DEFAULT_ACTION_IMPL,
	.execute = restart_service,
	.node = LIST_HEAD_INIT(service_restart_action.node),
};

FAULTD_ACTION_REGISTER_SIMPLE(service_restart_action);
