/*
 * This file is part of faultd.
 *
 * Copyright © 2017 Samsung Electronics
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "action.h"
#include "action_executor.h"
#include "decision_made_event.h"
#include "service_failed_event.h"
#include "resource_violation_event.h"
#include "log.h"
#include "systemd_dbus.h"

#ifdef FORCED_REBOOT_METHOD

#define DO_REBOOT(action, exe_info) reboot_forced(action, exe_info)

static int exit_handler(void)
{
	return system("/sbin/reboot");
}

static void reboot_forced(struct faultd_action *action,
					 struct action_executed_event *exe_info)
{
	sd_event *loop;
	int ret;

	loop = sd_event_source_get_event(action->execution_queue_sd_source);
	if (!loop) {
		faultd_object_append_string(exe_info->action_log, "error",
									"Failed to get event loop");
		ret = -1;
	} else {
		faultd_set_exit_handler(exit_handler);
		ret = sd_event_exit(loop, 0);
	}

	exe_info->result = ret;
}

#else

#define DO_REBOOT(action, exe_info) reboot_systemd(action, exe_info)

static void reboot_systemd(struct faultd_action *action,
						   struct action_executed_event *exe_info)
{
	int ret;

	/*
	 * We reboot the system without using logind because this action may be
	 * executed when for example dbus daemon is dead
	 */
	ret = faultd_dbus_call_systemd_simple(SYSTEMD_OBJ,
										  SYSTEMD_MANAGER_INTERFACE,
										  "StartUnit",
										  "ss",
										  "reboot.target",
										  "replace-irreversibly");
	if (ret < 0)
		faultd_object_append_string(exe_info->action_log, "error",
									"Failed to call systemd");

	exe_info->result = ret;
}

#endif

static int reboot_system(struct faultd_action *action,
						 struct action_executed_event *exe_info)
{
	struct faultd_event *ev = pop_faultd_event(&action->execution_queue);
	struct decision_made_event *dm_ev = to_decision_made_event(ev);
	struct service_failed_event *sf_ev = NULL;
	struct resource_violation_event *rv_ev = NULL;
	char *service_name = "unknown";

	if (faultd_event_is_of_type(dm_ev->reason,
								SERVICE_FAILED_EVENT_ID)) {

		sf_ev = to_service_failed_event(dm_ev->reason);
		service_name = sf_ev->service.dbus_path;

	} else if (faultd_event_is_of_type(dm_ev->reason,
									   RESOURCE_VIOLATION_EVENT_ID)) {

		rv_ev = to_resource_violation_event(dm_ev->reason);
		service_name = rv_ev->service.dbus_path;

	}

	log_kmsg("Trying to reboot the system (reason: %s)", service_name);

	/*
	 * We are passing an event to exe_info, so there is no need to unref it
	 * in this function
	 */
	exe_info->reason = ev;

	faultd_object_append_string(exe_info->action_log, FAULTD_AD_SERVICE_NAME,
								service_name);

	DO_REBOOT(action, exe_info);

	return 0;
}

static struct faultd_action system_reboot_action = {
	.action_id = FAULTD_ACTION_REBOOT_ID,
	.impl_name = FAULTD_DEFAULT_ACTION_IMPL,
	.execute = reboot_system,
	.node = LIST_HEAD_INIT(system_reboot_action.node),
};

FAULTD_ACTION_REGISTER_SIMPLE(system_reboot_action);
