/*
 * This file is part of faultd.
 *
 * Copyright © 2017 Samsung Electronics
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>

#include "action.h"
#include "action_executor.h"
#include "decision_made_event.h"
#include "log.h"
#include "systemd_dbus.h"

#ifdef FORCED_REBOOT_METHOD

#define DO_REBOOT(action, exe_info) reboot_forced(action, exe_info)

static int exit_handler(void)
{
	return system("/sbin/reboot recovery");
}

static void reboot_forced(struct faultd_action *action,
						 struct action_executed_event *exe_info)
{
	sd_event *loop;
	int ret;

	loop = sd_event_source_get_event(action->execution_queue_sd_source);
	if (!loop) {
		faultd_object_append_string(exe_info->action_log, "error",
									"Failed to get event loop");
		ret = -1;
	} else {
		faultd_set_exit_handler(exit_handler);
		ret = sd_event_exit(loop, 0);
	}

	exe_info->result = ret;
}

#else

#define REBOOT_PARAM_PATH "/run/systemd/reboot-param"
#define REBOOT_PARAM "recovery\n"
#define DO_REBOOT(action, exe_info) reboot_systemd(action, exe_info)

static void reboot_systemd(struct faultd_action *action,
						  struct action_executed_event *exe_info)
{
	mode_t prev_umask;
	FILE *rp_file;
	size_t n_wrte;
	int ret;

	prev_umask = umask(0022);

	rp_file = fopen(REBOOT_PARAM_PATH, "w");
	if (!rp_file) {
		faultd_object_append_string(exe_info->action_log, "fopen()",
									"Unable to open param file");
		goto reboot;
	}

	n_wrte = fwrite(REBOOT_PARAM, sizeof(REBOOT_PARAM), 1, rp_file);
	fclose(rp_file);
	umask(prev_umask);

	if (n_wrte < 1) {
		faultd_object_append_string(exe_info->action_log, "fwrite()",
									"Unable to set reboot param");
	}

reboot:
	/*
	 * We reboot the system without using logind because this action may be
	 * executed when for example dbus daemon is dead
	 */
	ret = faultd_dbus_call_systemd_simple(SYSTEMD_OBJ,
										  SYSTEMD_MANAGER_INTERFACE,
										  "StartUnit",
										  "ss",
										  "reboot.target",
										  "replace-irreversibly");
	if (ret < 0)
		faultd_object_append_string(exe_info->action_log, "error",
									"Failed to call systemd");

	exe_info->result = ret;
}

#endif

static int reboot_system_to_recovery(struct faultd_action *action,
									 struct action_executed_event *exe_info)
{
	struct faultd_event *ev = pop_faultd_event(&action->execution_queue);

	/*
	 * We are passing an event to exe_info, so there is no need to unref it
	 * in this function
	 */
	exe_info->reason = ev;

	log_kmsg("Trying to reboot to recovery");

	DO_REBOOT(action, exe_info);

	return 0;
}

static struct faultd_action system_reboot_to_recovery_action = {
	.action_id = FAULTD_ACTION_RECOVERY_REBOOT_ID,
	.impl_name = FAULTD_DEFAULT_ACTION_IMPL,
	.execute = reboot_system_to_recovery,
	.node = LIST_HEAD_INIT(system_reboot_to_recovery_action.node),
};

FAULTD_ACTION_REGISTER_SIMPLE(system_reboot_to_recovery_action);
