/*
 * This file is part of faultd.
 *
 * Copyright © 2017 Samsung Electronics
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <errno.h>
#include <systemd/sd-id128.h>

#include "common.h"
#include "event.h"
#include "log.h"
#include "module.h"

#include "system_booted_event.h"

static struct system_booted_event *boot_event = NULL;

void set_boot_event(struct system_booted_event *sb_ev)
{
	boot_event = sb_ev;
	faultd_event_ref(&boot_event->event);
}

struct system_booted_event *get_boot_event(void)
{
	if (boot_event)
		faultd_event_ref(&boot_event->event);
	return boot_event;
}

struct event_factory {
	struct faultd_module module;
	struct list_head types;
};

#define to_event_factory(MOD)						\
	container_of(MOD, struct event_factory, module)

static int init_event_factory(struct faultd_module *module,
							  struct faultd_config *config,
							  sd_event *event_loop)
{
	return 0;
}

static void cleanup_event_factory(struct faultd_module *module)
{
	struct event_factory *efactory = to_event_factory(module);

	if (boot_event != NULL)
		faultd_event_unref(&boot_event->event);

	assert(list_empty(&efactory->types));
}

static struct event_factory event_factory = {
	.module = {
		.name = "event_factory",
		.type = FAULTD_MODULE_TYPE_CORE,

		.init = init_event_factory,
		.cleanup = cleanup_event_factory,
		.node = LIST_HEAD_INIT(event_factory.module.node),
	},
	.types = LIST_HEAD_INIT(event_factory.types),
};

FAULTD_MODULE_REGISTER(&event_factory.module);

int faultd_event_type_register(struct faultd_event_type *new)
{
	if (!new->name || !new->allocate_event)
		return -EINVAL;

	list_add_tail(&new->node, &event_factory.types);

	return 0;
}

void faultd_event_type_unregister(struct faultd_event_type *type)
{
	if (list_empty(&type->node))
		return;

	list_del_init(&type->node);
}

int faultd_event_create(const char *type, void *data, struct faultd_event **ev)
{
	struct faultd_event_type *ev_type;
	bool from_bson;
	int ret = -EINVAL;

	if (!type)
		return ret;

	from_bson = (strcmp(type, BSON_EVENT_ID) == 0);
	if (from_bson) {
		struct faultd_object *d = data;
		/* get event type name from bson */
		ret = faultd_object_get_string(d, EV_TYPE_NAME, (char **)&type);
		if (ret < 0)
			return ret;
	}

	ret = -ENOENT;
	list_for_each_entry(ev_type, &event_factory.types, node)
		if (strcmp(ev_type->name, type) == 0)
			break;
	if (&ev_type->node == &event_factory.types)
		return ret;

	if (from_bson) {
		if (ev_type->deserialize_event == NULL) {
			ret = -ENOTSUP;
			log_error("Event type \"%s\" does not support deserialisation.",
					  ev_type->name);
		} else {
			ret = ev_type->deserialize_event(ev_type, data, ev);
			/* faultd_event_init_internal here? (old boot_id and oid) */
		}
	} else {
		ret = ev_type->allocate_event(ev_type, data, ev);
		/* faultd_event_init_internal here? (new boot_id and oid) */
	}
	return ret;
}

static void release_faultd_event(struct uref *uref)
{
	struct faultd_event *ev = container_of(uref, struct faultd_event, uref);

	if (ev->boot_event)
		faultd_event_unref(&ev->boot_event->event);

	if (ev->ops.release)
		ev->ops.release(ev);
}

char *faultd_event_to_string(struct faultd_event *ev)
{
	struct faultd_object *obj = NULL;
	char *str = NULL;
	int ret;

	ret = faultd_object_new(&obj);
	if (ret) {
		log_error("Unable to allocate faultd object");
		return NULL;
	}

	ev->ops.serialize(ev, obj);

	ret = faultd_object_to_string(&str, obj, 0);
	faultd_object_unref(obj);
	if (ret) {
		log_error("Unable to convert faultd object to string");
		return NULL;
	}

	return str;
}

int faultd_event_init_internal(struct faultd_event_type *ev_type,
							   struct faultd_event *ev)
{
	int ret;

	ret = clock_gettime(CLOCK_MONOTONIC, &ev->timestamp);
	if (ret)
		return ret;

	ev->boot_event = get_boot_event();

	ev->type = ev_type;
	ev->ops = ev_type->default_ops;

	uref_init(&ev->uref, release_faultd_event);
	INIT_LIST_HEAD(&ev->node);
	INIT_NQUEUE_NODE(&ev->event_queue_node.node);
	ev->event_queue_node.ev = ev;

	return 0;
}

void faultd_event_serialize_internal(struct faultd_event *ev, struct faultd_object *out)
{
	/* FIXME: check return values of faultd_object_append_XXX */
	if (ev->boot_event != NULL)
		faultd_object_append_oid(out, EV_BOOT_EVENT, &ev->boot_event->event.oid);
	/* FIXME: serialise event_type structure (?) */
	faultd_object_append_string(out, EV_TYPE_NAME, ev->type->name);
	faultd_object_append_timespec(out, EV_TIMESTAMP, &ev->timestamp);
}

int faultd_event_deserialize_internal(const struct faultd_object* in,
									  struct faultd_event_type *ev_type,
									  struct faultd_event *ev)
{
	struct faultd_object *obj;
	struct system_booted_event *current_boot_event = get_boot_event();
	struct faultd_event *other_boot_event = NULL;
	int ret = 0;

	ev->type = ev_type;
	ev->ops = ev_type->default_ops;
	ev->boot_event = NULL;

	list_for_each_entry(obj, &in->val.children, node)
		if (obj->type == TYPE_OID && strcmp(EV_BOOT_EVENT, obj->key) == 0) {
			if (current_boot_event &&
				faultd_oid_equal(&obj->val.oid, &current_boot_event->event.oid)) {

				/* event created in the current faultd instance */
				ev->boot_event = current_boot_event;
			} else {

				/* retrieve event's boot_event from database */
				ret = database_get_event(&obj->val.oid, &other_boot_event);
				if (ret == 0)
					ev->boot_event = to_system_booted_event(other_boot_event);
				else
					goto cleanup;
			}
		} else if (obj->type == TYPE_TIMESPEC && strcmp(EV_TIMESTAMP, obj->key) == 0) {
				memcpy(&ev->timestamp, &obj->val.ts, sizeof(ev->timestamp));
		}

cleanup:
	if (current_boot_event && ev->boot_event != current_boot_event)
		faultd_event_unref(&current_boot_event->event);

	return ret;
}

int append_faultd_event(struct nqueue_head *head, struct faultd_event *ev)
{
	struct eq_node *event_queue_node;

	if (list_empty(&ev->event_queue_node.node)) {
		event_queue_node = &ev->event_queue_node;
	} else {
		event_queue_node = malloc(sizeof(*event_queue_node));
		if (!event_queue_node) {
			log_error("Unable to allocate memory");
			return -ENOMEM;
		}
		event_queue_node->ev = ev;
	}

	faultd_event_ref(ev);

	return nqueue_append(head, &event_queue_node->node);
}

struct faultd_event *pop_faultd_event(struct nqueue_head *head)
{
	struct eq_node *event_queue_node = nqueue_pop(head,
												  struct eq_node,
												  node);
	struct faultd_event *ev;

	if (!event_queue_node) {
		log_error("Could not pop event queue");
		return NULL;
	}

	ev = event_queue_node->ev;

	if (event_queue_node == &ev->event_queue_node)
		INIT_NQUEUE_NODE(&ev->event_queue_node.node);
	else
		free(event_queue_node);

	return ev;
}
