/*
 * This file is part of faultd.
 *
 * Copyright © 2017 Samsung Electronics
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef FAULTD_EVENT_H
#define FAULTD_EVENT_H

#include <string.h>
#include <systemd/sd-id128.h>
#include <time.h>

#include "uref.h"
#include "list.h"
#include "notify_queue.h"
#include "module.h"

#define EV_BOOT_EVENT "bev"
#define EV_TIMESTAMP "ts"
#define EV_TYPE_NAME "tn"

#define BSON_EVENT_ID "bson_event"

struct faultd_event;

struct faultd_event_ops {
	void (*release)(struct faultd_event *);
	void (*serialize)(struct faultd_event*, struct faultd_object*);
};

struct faultd_event_type {
	char *name;

	struct faultd_event_ops default_ops;

	/* Add function to allocate event from DB */

	/* Allocate event based on passed data */
	int (*allocate_event)(struct faultd_event_type *type,
						  void *data, struct faultd_event **ev);

	int (*deserialize_event)(struct faultd_event_type *type,
							 struct faultd_object *data, struct faultd_event **ev);

	/* To be used by event factory */
	struct list_head node;
};

struct eq_node {
	struct faultd_event *ev;
	struct nqueue_node node;
};

struct faultd_event {
	struct timespec timestamp;
	struct system_booted_event *boot_event;

	/* TODO: add here some id field */

	struct faultd_event_type *type;
	struct faultd_event_ops ops;

	struct uref uref;
	/* To be used by event holder */
	struct list_head node;
	/* To be used by event processor */
	struct eq_node event_queue_node;

	faultd_oid_t oid;
};

int faultd_event_type_register(struct faultd_event_type *type);
void faultd_event_type_unregister(struct faultd_event_type *type);

#define FAULTD_EVENT_TYPE_REGISTER(EVENT_TYPE, NAME)				\
static int CONCAT(CONCAT(event_type_, FAULTD_MODNAME_T), _init)(struct faultd_module *module, \
									  struct faultd_config *config, \
									  sd_event *event_loop)			\
{																	\
	return faultd_event_type_register(&EVENT_TYPE);                 \
}																	\
																	\
static void CONCAT(CONCAT(event_type_, FAULTD_MODNAME_T), _cleanup)(struct faultd_module *module) \
{																	\
	faultd_event_type_unregister(&EVENT_TYPE);                      \
}																	\
																	\
static struct faultd_module CONCAT(CONCAT(event_type_, FAULTD_MODNAME_T), _module) = {		\
	.name = FAULTD_MODNAME,											\
	.type = FAULTD_MODULE_TYPE_EVENT,                               \
	.init = CONCAT(CONCAT(event_type_, FAULTD_MODNAME_T), _init),	\
	.cleanup = CONCAT(CONCAT(event_type_, FAULTD_MODNAME_T), _cleanup), \
	.node = LIST_HEAD_INIT(CONCAT(CONCAT(event_type_, FAULTD_MODNAME_T), _module.node)), \
};																	\
																	\
FAULTD_MODULE_REGISTER(&CONCAT(CONCAT(event_type_, FAULTD_MODNAME_T), _module))

int faultd_event_create(const char *type, void *data, struct faultd_event **ev);

static inline void faultd_event_ref(struct faultd_event *ev)
{
	uref_get(&ev->uref);
}

static inline void faultd_event_unref(struct faultd_event *ev)
{
	uref_put(&ev->uref);
}

static inline int faultd_event_is_of_type(struct faultd_event *ev,
											  const char *type)
{
	return strcmp(ev->type->name, type) == 0;
}

char *faultd_event_to_string(struct faultd_event *ev);

static inline void faultd_event_serialize(struct faultd_event *ev, struct faultd_object *out)
{
	ev->ops.serialize(ev, out);
}

int faultd_event_init_internal(struct faultd_event_type *ev_type,
							   struct faultd_event *ev);

static inline void faultd_event_cleanup_internal(struct faultd_event *ev) {}

void faultd_event_serialize_internal(struct faultd_event *ev, struct faultd_object *out);

int faultd_event_deserialize_internal(const struct faultd_object *in, struct faultd_event_type *ev_type, struct faultd_event *ev);

void set_boot_event(struct system_booted_event *sb_ev);
struct system_booted_event *get_boot_event(void);

int append_faultd_event(struct nqueue_head *head, struct faultd_event *ev);
struct faultd_event *pop_faultd_event(struct nqueue_head *head);

#endif /* FAULTD_EVENT_H */
