/*
 * This file is part of faultd.
 *
 * Copyright © 2017 Samsung Electronics
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <json-c/json.h>

#include "common.h"
#include "faultd-config.h"
#include "json-config.h"
#include "log.h"
#include "module.h"

static struct faultd_config modules_config;
static struct faultd_config services_config;

void faultd_get_modules_config(struct faultd_config **c)
{
	*c = &modules_config;
}

void faultd_get_services_config(struct faultd_config **c)
{
	*c = &services_config;
}

static int parse_disable_modules(void)
{
	struct json_object *root, *node;
	const char *name;
	struct faultd_module *module;
	int len, ret;

	ret = json_object_object_get_ex(modules_config.root, "disable_modules", &root);
	if (ret == 0)
		return 0;

	len = json_object_array_length(root);

	for (int i = 0; i < len; ++i) {
		node = json_object_array_get_idx(root, i);
		if (!node)
			return -EINVAL;

		name = json_object_get_string(node);
		if (!name)
			return -EINVAL;

		module = faultd_get_module_by_name(name);
		if (module) {
			module->disabled = 1;
			log_debug("Disabled %s module", name);
		} else {
			log_warning("Could not find %s module", name);
		}
	}

	return 0;
}

static int init_config(void)
{
	modules_config.root = json_object_new_object();
	if (!modules_config.root) {
			log_error("Unable to create new json object");
			return -ENOMEM;
	}

	services_config.root = json_object_new_object();
	if (!services_config.root) {
			log_error("Unable to create new json object");
			json_object_put(modules_config.root);
			return -ENOMEM;
	}

	return 0;
}

static void log_config(void)
{
	log_debug("Modules config root:\n\n%s\n",
			  json_object_to_json_string_ext(modules_config.root, JSON_C_TO_STRING_PRETTY));

	log_debug("Services config root:\n\n%s\n",
			  json_object_to_json_string_ext(services_config.root, JSON_C_TO_STRING_PRETTY));
}

int faultd_load_config(const char *custom_config_file)
{
	int ret;

	ret = init_config();
	if (ret)
		return ret;

	ret = load_config(FAULTD_DEFAULT_CONFIG_FILE, CONFIG_TYPE_FILE, 1,
					  modules_config.root);
	if (ret)
		goto cleanup;

	ret = load_config(FAULTD_DEFAULT_MODULES_CONFIG_PATH, CONFIG_TYPE_DIR, 0,
					  modules_config.root);
	if (ret)
		goto cleanup;

	ret = load_config(FAULTD_DEFAULT_SERVICES_CONFIG_PATH, CONFIG_TYPE_DIR, 0,
					  services_config.root);
	if (ret)
		goto cleanup;

	ret = load_config(custom_config_file, CONFIG_TYPE_FILE, 1,
					  modules_config.root);
	if (ret)
		goto cleanup;

	ret = load_config(FAULTD_CUSTOM_MODULES_CONFIG_PATH, CONFIG_TYPE_DIR, 1,
					  modules_config.root);
	if (ret)
		goto cleanup;

	ret = load_config(FAULTD_CUSTOM_SERVICES_CONFIG_PATH, CONFIG_TYPE_DIR, 1,
					  services_config.root);
	if (ret)
		goto cleanup;

	log_config();

	ret = parse_disable_modules();
	if (ret) {
		log_error("Could not parse disable_modules section");
		goto cleanup;
	}

	return 0;

cleanup:
	json_object_put(modules_config.root);
	json_object_put(services_config.root);
	return ret;
}

void faultd_config_cleanup(struct faultd_config *c)
{
	if (c)
		json_object_put(c->root);
}

void faultd_config_cleanup_global(void)
{
	faultd_config_cleanup(&modules_config);
	faultd_config_cleanup(&services_config);
}

int faultd_config_get_subconf(struct faultd_config *conf, const char *key, struct faultd_config **sub)
{
	struct faultd_config *config_node;
	struct json_object *node;
	json_bool ret;

	ret = json_object_object_get_ex(conf->root, key, &node);
	if (!ret)
		return -ENOENT;

	config_node = calloc(1, sizeof(*config_node));
	if (!config_node)
		return -ENOMEM;

	if (!json_object_is_type(node, json_type_object)) {
		log_error("%s has to be an object", key);
		goto out;
	}

	config_node->root = json_object_get(node);

	*sub = config_node;
	return 0;

out:
	free(config_node);
	return -EINVAL;
}
