/*
 * This file is part of faultd.
 *
 * Copyright © 2017 Samsung Electronics
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <glib.h>
#include <systemd/sd-event.h>
#include <errno.h>

#include "common.h"
#include "faultd-glib.h"

struct faultd_event_source {
	GSource gsource; /* gsource should be the first field of this structure */
	sd_event *event;
};

#define to_faultd_event_source(source) \
	container_of(source, struct faultd_event_source, gsource)

static gboolean prepare(GSource *source, gint *timeout)
{
	struct faultd_event_source *s = to_faultd_event_source(source);

	return sd_event_prepare(s->event);
}

static gboolean check(GSource *source)
{
	struct faultd_event_source *s = to_faultd_event_source(source);

	return sd_event_wait(s->event, 0) > 0;
}

static gboolean dispatch(GSource *source, GSourceFunc callback, gpointer user_data)
{
	struct faultd_event_source *s = to_faultd_event_source(source);

	return sd_event_dispatch(s->event);
}

static void finalize(GSource *source)
{
	struct faultd_event_source *s = to_faultd_event_source(source);

	sd_event_unref(s->event);
}

static void faultd_on_exit(sd_event_source *s, void *userdata)
{
	GMainLoop *gloop = userdata;

	g_main_loop_quit(gloop);
}

static GSourceFuncs source_funcs = {
	.prepare = prepare,
	.check = check,
	.dispatch = dispatch,
	.finalize = finalize,
};

int faultd_g_main_loop_wrapper(sd_event *loop)
{
	GMainLoop *gloop;
	GSource *source;
	GPollFD pollfd;
	struct faultd_event_source *s;
	int ret;

	gloop = g_main_loop_new(NULL, FALSE);
	if (!gloop)
		return -ENOMEM;

	source = g_source_new(&source_funcs, sizeof(struct faultd_event_source));
	if (!source) {
		ret = -ENOMEM;
		goto gloop_unref;
	}

	s = to_faultd_event_source(source);
	s->event = sd_event_ref(loop);
	sd_event_add_exit(loop, NULL, faultd_on_exit, gloop);

	pollfd.fd = sd_event_get_fd(loop);
	pollfd.events = G_IO_IN | G_IO_HUP | G_IO_ERR;
	g_source_add_poll(source, &pollfd);

	g_source_attach(source, NULL);

	g_main_loop_run(gloop);
	ret = 0;

	g_source_unref(source);

gloop_unref:
	g_main_loop_unref(gloop);

	return ret;
}
