/*
 * This file is a part of fauld.
 *
 * Copyright © 2017 Samsung Electronics
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#include <assert.h>
#include <ejdb/ejdb.h>
#include <errno.h>
#include <systemd/sd-id128.h>

#include "database.h"
#include "log.h"
#include "module.h"

struct ejdb_adapter {
	struct faultd_database_adapter database_adapter;
	EJDB *db;
	EJCOLL *coll;
};

#define to_ejdb_adapter(MOD)				\
	container_of(MOD, struct ejdb_adapter, database_adapter)

static void cleanup_ejdb_adapter(struct faultd_module *module)
{
	struct faultd_database_adapter *_da = to_database_adapter(module);
	struct ejdb_adapter *da = to_ejdb_adapter(_da);

	if (da->db == NULL)
		return;

	da->coll = NULL;

	if (!ejdbclose(da->db))
		log_error("ejdbclose has failed: %s", ejdberrmsg(ejdbecode(da->db)));

	ejdbdel(da->db);
	da->db = NULL;
}

static int init_ejdb_adapter(struct faultd_module *module,
								 struct faultd_config *config,
								 sd_event *event_loop)
{
	struct faultd_database_adapter *_da = to_database_adapter(module);
	struct ejdb_adapter *da = to_ejdb_adapter(_da);
	int ret;

	ret = faultd_set_database_adapter(_da);
	if (ret < 0) {
		log_error("Database adapter was already set. Please, enable only one database module");
		return -1;
	}

	da->db = ejdbnew();
	if (da->db == NULL) {
		log_error("ejdbnew has failed: %m");
		goto error;
	}

	/* when faultdb_log.wal file exists, ejdbopen() will cause segfault */
	ret = access(DBDIR "/faultdb_log.wal", F_OK);
	if (ret == 0) {
		log_error("Database was not closed properly. "
				  "Removing faultdb_log.wal before opening database again");

		ret = unlink(DBDIR "/faultdb_log.wal");
		if (ret != 0) {
			log_error_errno(errno, "Unable to remove faultdb_log.wal");
			goto error;
		}
	}

	if (!ejdbopen(da->db, DBDIR "/faultdb", JBOWRITER | JBOCREAT | JBOTSYNC)) {
		log_error("ejdbopen has failed: %s", ejdberrmsg(ejdbecode(da->db)));
		goto error;
	}

	da->coll = ejdbcreatecoll(da->db, "log", NULL);
	if (da->coll == NULL) {
		log_error("ejdbcreatecoll has failed: %s", ejdberrmsg(ejdbecode(da->db)));
		goto error;
	}
	return 0;

error:
	cleanup_ejdb_adapter(module);
	return -1;
}

static int faultd_object_to_bson(struct faultd_object *obj, bson *out)
{
	struct faultd_object *child;

	if (obj->type != TYPE_OBJECT)
		return -EINVAL;

	if (obj->flags & FAULTD_FLAG_QUERY_MODE)
		bson_init_as_query(out);
	else
		bson_init(out);

	list_for_each_entry(child, &obj->val.children, node) {
		switch (child->type) {
		case TYPE_OID:
			bson_append_oid(out, child->key, &child->val.oid.bson);
			break;
		case TYPE_STRING:
			bson_append_string(out, child->key, child->val.s);
			break;
		case TYPE_INT:
			bson_append_int(out, child->key, child->val.i);
			break;
		case TYPE_TIMESPEC:
			{
				bson_append_start_object(out, child->key);
				bson_append_int(out, "tv_sec", child->val.ts.tv_sec);
				bson_append_long(out, "tv_nsec", child->val.ts.tv_nsec);
				bson_append_finish_object(out);
				break;
			}
		case TYPE_TIME_T:
			bson_append_time_t(out, child->key, child->val.time);
			break;
		case TYPE_UUID:
			bson_append_binary(out, child->key, BSON_BIN_UUID,
					(char *)(&child->val.uuid), sizeof(child->val.uuid));
			break;
		case TYPE_OBJECT:
			{
				bson b;

				faultd_object_to_bson(child, &b);
				bson_append_bson(out, child->key, &b);
				bson_destroy(&b);
				break;
			}
		}
	}

	bson_finish(out);
	return 0;
}
TEST_ENTRY_POINT(faultd_object_to_bson);

static int bson_to_timespec(bson *in, struct timespec *ts)
{
	bson_iterator it;

	BSON_ITERATOR_INIT(&it, in);

	if ((bson_iterator_next(&it) == BSON_INT) &&
		(strcmp(bson_iterator_key(&it), "tv_sec") == 0))

		ts->tv_sec = bson_iterator_int(&it);
	else
		return -ENOENT;

	if ((bson_iterator_next(&it) == BSON_LONG) &&
		(strcmp(bson_iterator_key(&it), "tv_nsec") == 0))

		ts->tv_nsec = bson_iterator_long(&it);
	else
		return -ENOENT;

	if (bson_iterator_next(&it) != BSON_EOO)
		return -ENOENT;

	return 0;
}

static int bson_to_faultd_object(bson *b, struct faultd_object *out)
{
	bson_type bt;
	bson_iterator it;
	int ret;

	BSON_ITERATOR_INIT(&it, b);

	while ((bt = bson_iterator_next(&it)) != BSON_EOO) {
		const char *key = bson_iterator_key(&it);
		faultd_oid_t oid;

		switch (bt) {
		case BSON_OID:
			memset(&oid, 0, sizeof(oid));
			oid.bson = *bson_iterator_oid(&it);
			ret = faultd_object_append_oid(out, key, &oid);
			break;
		case BSON_STRING:
			ret = faultd_object_append_string(out, key, bson_iterator_string(&it));
			break;
		case BSON_INT:
			ret = faultd_object_append_int(out, key, bson_iterator_int(&it));
			break;
		case BSON_LONG:
			ret = faultd_object_append_time_t(out, key, bson_iterator_time_t(&it));
			break;
		case BSON_DATE:
			ret = faultd_object_append_time_t(out, key, bson_iterator_time_t(&it));
			break;
		case BSON_BINDATA:
			switch (bson_iterator_bin_type(&it)) {
			case BSON_BIN_UUID:
				ret = faultd_object_append_uuid(out, key,
					  (sd_id128_t *)bson_iterator_bin_data(&it));
				break;
			default:
				ret = -ENOTSUP;
				break;
			}
			break;
		case BSON_OBJECT:
			{
				struct timespec ts;
				bson bs;

				bson_iterator_subobject(&it, &bs);

				if (bson_to_timespec(&bs, &ts) == 0) {
					ret = faultd_object_append_timespec(out, key, &ts);
				} else {
					struct faultd_object *child;

					ret = faultd_object_new(&child);
					if (ret != 0)
						break;

					ret = bson_to_faultd_object(&bs, child);
					if (ret != 0)
						goto cleanup;

					ret = faultd_object_append_object(out, key, child);

				cleanup:
					faultd_object_unref(child);
				}

				break;
			}

		default:
			ret = -ENOTSUP;
		}

		if (ret < 0)
			return ret;
	}

	return 0;
}
TEST_ENTRY_POINT(bson_to_faultd_object);

static int ejdb_store(struct faultd_database_adapter *adapter, struct faultd_object *obj, faultd_oid_t *oid)
{
	struct ejdb_adapter *da = to_ejdb_adapter(adapter);
	EJDB *db = da->db;
	EJCOLL *coll = da->coll;
	bson b;

	if (!ejdbtranbegin(coll)) {
		log_error("ejdbtransbegin failed: %s", ejdberrmsg(ejdbecode(db)));
		return -1;
	}

	if (faultd_object_to_bson(obj, &b) < 0)
		goto abort;

	if (!ejdbsavebson(coll, &b, &oid->bson)) {
		log_error("ejdbsavebson failed: %s", ejdberrmsg(ejdbecode(db)));
		goto destroy_bson;
	}

	if (!ejdbtrancommit(coll)) {
		log_error("ejdbtrancommit failed: %s", ejdberrmsg(ejdbecode(db)));
		goto destroy_bson;
	}

	bson_destroy(&b);

	return 0;

destroy_bson:
	bson_destroy(&b);

abort:
	if (!ejdbtranabort(coll))
		/* That's BAD! */
		log_error("ejdbtranabort failed: %s", ejdberrmsg(ejdbecode(db)));
	return -1;
}

static int ejdb_get_by_oid(struct faultd_database_adapter *adapter, faultd_oid_t *oid, struct faultd_object *result)
{
	struct ejdb_adapter *da = to_ejdb_adapter(adapter);
	EJCOLL *coll = da->coll;
	bson *b;
	int ret;

	b = ejdbloadbson(coll, &oid->bson);
	if (b == NULL)
		return -ENOENT;

	ret = bson_to_faultd_object(b, result);
	bson_del(b);
	return ret;
}

static int ejdb_load(struct faultd_database_adapter *adapter, struct faultd_object *query,
					 struct faultd_object *hints, struct faultd_object *result, uint32_t *nr)
{
	struct ejdb_adapter *da = to_ejdb_adapter(adapter);
	EJDB *db = da->db;
	EJCOLL *coll = da->coll;
	EJQ *q = NULL;
	TCLIST *r = NULL;
	uint32_t count = 0;
	int ret = -EINVAL;
	int qflags = 0;
	bson bq;
	bson *bhints = NULL;

	if (result == NULL)
		qflags |= JBQRYCOUNT;

	ret = faultd_object_to_bson(query, &bq);
	if (ret < 0)
		return ret;

	if (hints) {
		bhints = bson_create();
		if (bhints == NULL) {
			log_error("Unable to create bson object");
			goto finish;
		}
		ret = faultd_object_to_bson(hints, bhints);
		if (ret < 0)
			goto finish;
	}

	q = ejdbcreatequery(db, &bq, NULL, 0, bhints);

	if (q == NULL) {
		log_error("ejdbcreatequery failed: %s", ejdberrmsg(ejdbecode(db)));
		goto finish;
	}

	r = ejdbqryexecute(coll, q, &count, qflags, NULL);
	if ((r == NULL) && (ejdbecode(db) != TCESUCCESS)) {
		log_error("ejdbqryexecute failed: %s", ejdberrmsg(ejdbecode(db)));
		goto finish;
	}

	log_debug("ejdbqueryexecute: %u records", count);

	if (r == NULL) {
		ret = 0;
		goto finish;
	}

	for (uint32_t i = 0; i < count; ++i) {
		void *bsdata = TCLISTVALPTR(r, i);
		bson val;
		struct faultd_object *obj;
		char key[11]; /* strlen(itoa(1U<<31, key, 10)) => 10 */

		bson_init_finished_data(&val, bsdata);
		snprintf(key, sizeof(key), "%u", i);
		ret = faultd_object_new(&obj);
		if (ret < 0)
			goto finish;

		bson_to_faultd_object(&val, obj);
		/* TODO Check if binary fields are matching if there are any*/
		faultd_object_append_object(result, key, obj);
		faultd_object_unref(obj);
	}

	ret = 0;
finish:
	if (bhints != NULL)
		bson_del(bhints);

	bson_destroy(&bq);

	if (r != NULL)
		tclistdel(r);

	if (q != NULL)
		ejdbquerydel(q);

	if (ret == 0) {
		if (nr != NULL)
			*nr = count;
	}

	return ret;
}

static int ejdb_get_well_known_oid(const char *name, faultd_oid_t *oid)
{
	int i;
	static const struct {
		const char *name;
		const char *oid;
	} oid_strings[] = {
		{FAULTD_LAST_BOOT_OID, "32a4ca4f96ddfcd8a1f50269014ff320"},
		{NULL, NULL}
	};

	for (i = 0; oid_strings[i].name; ++i) {
		if (strcmp(oid_strings[i].name, name))
			continue;

		bson_oid_from_string(&oid->bson, oid_strings[i].oid);
		return 0;
	}

	return -ENOENT;
}

static bool ejdb_is_oid_valid(faultd_oid_t *oid)
{
	return oid->bson.ints[0] != 0 ||
		oid->bson.ints[1] != 0 ||
		oid->bson.ints[2] != 0;
}

static struct ejdb_adapter ejdb_adapter = {
	.database_adapter = {
		.module = {
			.name = "ejdb_dbadapter",
			.type = FAULTD_MODULE_TYPE_DBADAPTER,
			.init = init_ejdb_adapter,
			.cleanup = cleanup_ejdb_adapter,
			.node = LIST_HEAD_INIT(ejdb_adapter.database_adapter.module.node),
		},
		.id_key = JDBIDKEYNAME,
		.store = ejdb_store,
		.get_by_oid = ejdb_get_by_oid,
		.load = ejdb_load,
		.get_well_known_oid = ejdb_get_well_known_oid,
		.is_oid_valid = ejdb_is_oid_valid,
	}
};

FAULTD_MODULE_REGISTER(&ejdb_adapter.database_adapter.module);
