/*
 * This file is a part of faultd.
 *
 * Copyright © 2017 Samsung Electronics
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#include <assert.h>
#include <errno.h>
#include <sqlite3.h>

#include "database.h"
#include "log.h"
#include "module.h"

#define META_TABLE "meta"
#define OBJ_TABLE_PREFIX "objects"
#define ID_KEY "id"
#define BUFSIZE 512
#define WK_TABLE_IDX 0

enum well_known_oids {
	WK_OID_INVALID = 0,
	WK_LAST_BOOT,
	WK_OID_END,
};

const char *well_known_signatures[WK_OID_END] = {
	[WK_LAST_BOOT] = ":{boot_event:o}",
};

enum prepared_stmts {
	STMT_GET_ALL_TABLES = 0,
	STMT_GET_TABLE_BY_ID,
	STMT_FIND_TABLE,
	STMT_INSERT_TABLE,
	STMT_NUM,
};

const char *prepared_stmts_str[STMT_NUM] = {
	[STMT_GET_ALL_TABLES] = "SELECT ('" OBJ_TABLE_PREFIX "'||id), signature FROM " META_TABLE ";",
	[STMT_GET_TABLE_BY_ID] = "SELECT ('" OBJ_TABLE_PREFIX "'||id), signature FROM " META_TABLE " where id = ?;",
	[STMT_FIND_TABLE] = "SELECT ('" OBJ_TABLE_PREFIX "'||id), id FROM " META_TABLE " WHERE signature = ?;",
	[STMT_INSERT_TABLE] = "INSERT INTO " META_TABLE " (signature) VALUES (?);",
};

struct sqlite_adapter {
	struct faultd_database_adapter database_adapter;
	sqlite3 *db;
	sqlite3_stmt *stmts[STMT_NUM];
};

#define to_sqlite_adapter(MOD)				\
	container_of(MOD, struct sqlite_adapter, database_adapter)

static void cleanup_sqlite_adapter(struct faultd_module *module)
{
	struct faultd_database_adapter *_da = to_database_adapter(module);
	struct sqlite_adapter *da = to_sqlite_adapter(_da);
	int i;

	for (i = 0; i < STMT_NUM; ++i)
		sqlite3_finalize(da->stmts[i]);

	sqlite3_close(da->db);
}

static int init_sqlite_adapter(struct faultd_module *module,
								 struct faultd_config *config,
								 sd_event *event_loop)
{
	struct faultd_database_adapter *_da = to_database_adapter(module);
	struct sqlite_adapter *da = to_sqlite_adapter(_da);
	int ret;
	char *sqlite_err = NULL;
	int prepared = 0;

	ret = sqlite3_open(DBDIR "/faultd.sqlite", &da->db);
	if (ret < 0) {
		log_error("Could not open database: %s\n", sqlite3_errmsg(da->db));
		return ret;
	}

	ret = sqlite3_exec(da->db, "CREATE TABLE IF NOT EXISTS " META_TABLE
							   "(id INTEGER PRIMARY KEY, signature TEXT)"
							   ";", NULL, 0, &sqlite_err);
	if (ret < 0) {
		log_error("SQL Error: %s\n", sqlite3_errmsg(da->db));
		sqlite3_free(sqlite_err);
		goto close_db;
	}

	for (prepared = 0; prepared < STMT_NUM; ++prepared) {
		ret = sqlite3_prepare_v2(da->db, prepared_stmts_str[prepared], -1, &da->stmts[prepared], NULL);
		if (ret < 0) {
			log_error_errno(ret, "Could not prepare sql statement:\n %s", prepared_stmts_str[prepared]);
			goto out;
		}
	}

	ret = faultd_set_database_adapter(_da);
	if (ret < 0) {
		log_error("Database adapter was already set. Please, enable only one database module");
		goto out;
	}

	return 0;

out:
	while (prepared--)
		sqlite3_finalize(da->stmts[prepared]);

close_db:
	sqlite3_close(da->db);
	return ret;
}

#define OBJ_SIGNATURE_BEGIN '{'
#define OBJ_SIGNATURE_END '}'

static char type_id_map[] = {
	[TYPE_OBJECT] = OBJ_SIGNATURE_BEGIN,
	[TYPE_STRING] = 's',
	[TYPE_OID] = 'o',
	[TYPE_INT] = 'i',
	[TYPE_TIMESPEC] = 'T',
	[TYPE_TIME_T] = 't',
	[TYPE_UUID] = 'u',
};

static const char *type_sql_map[] = {
	[TYPE_OBJECT] = "NULL",
	[TYPE_STRING] = "TEXT",
	[TYPE_OID] = "INTEGER",
	[TYPE_INT] = "INTEGER",
	[TYPE_TIMESPEC] = "BLOB",
	[TYPE_TIME_T] = "INTEGER",
	[TYPE_UUID] = "BLOB",
};

enum signature_type {
	SIG_KEY, /**< Signature stored in database */
	SIG_CREATE_FIELDS, /**< Fields for create query */
	SIG_INSERT_FIELDS, /**< Fields for insert query */
	SIG_INSERT_VALUES, /**< Values for insert query */
	SIG_LIKE_PATTERN, /**< Pattern for sql LIKE operator */
	SIG_WHERE_QUERY, /**< WHERE operators for query */
};

static const char *get_query_op(const char *key)
{
	int i;
	static const struct {
		const char *k;
		const char *v;
	} map[] = {
		{"$not", "!="},
		{"$lt", "<"},
		{"$lte", "<="},
		{"$gte", ">"},
		{"$gt", ">="},
		{NULL, NULL},
	};

	for (i = 0; map[i].k; ++i)
		if (strcmp(map[i].k, key) == 0)
			return map[i].v;

	return NULL;
}

static int _get_object_signature(struct faultd_object *obj, char *buffer, size_t size, int offset, int type, const char *key_prefix)
{
	struct faultd_object *child;
	int ret = 0;
	int noff = offset;
	char *prefix = NULL;
	const char *op;

	if (obj->key && (strcmp(obj->key, ID_KEY) == 0
				&& type != SIG_INSERT_VALUES && type != SIG_INSERT_FIELDS
				|| ((type != SIG_WHERE_QUERY && obj->key[0] == '$'))))
		return 0;

	if (obj->type == TYPE_OBJECT) {
		if (type == SIG_KEY) {
			ret = snprintf(buffer + noff, size - noff, "%s:{", (obj->key ? obj->key : ""));
			if (ret < 0 || ret >= (int)size - noff)
				return -ENOMEM;

			noff += ret;
		}

		if (obj->key)
			ret = asprintf(&prefix, "%s%s.", (key_prefix ? key_prefix : ""), obj->key);
		else
			prefix = strdup(key_prefix);

		if (ret < 0 || !prefix)
			return -ENOMEM;

		list_for_each_entry(child, &obj->val.children, node) {
			if (type == SIG_LIKE_PATTERN)
				buffer[noff++] = '%';

			ret = _get_object_signature(child, buffer, size, noff, type, prefix);
			if (ret < 0)
				goto out;

			noff += ret;
		}

		free(prefix);
		prefix = NULL;

		if (type == SIG_KEY) {
			if (buffer[noff - 1] == ',')
				noff--;

			ret = snprintf(buffer + noff, size - noff, "},");
			if (ret < 0 || ret >= (int)size - noff)
				return -ENOMEM;

			noff += ret;
		}

		ret = noff - offset;
		if (ret < 0)
			return 0;

		return ret;
	}

	switch (type) {
	case SIG_KEY:
	case SIG_LIKE_PATTERN:
		if (obj->key && size < offset + strlen(obj->key) + 3)
			return -ENOMEM;

		ret = snprintf(buffer + offset, size - offset,
				"%s:%c,", (obj->key ? obj->key : ""), type_id_map[obj->type]);
		break;
	case SIG_CREATE_FIELDS:
		ret = snprintf(buffer + offset, size - offset,
				"`%s%s` %s,", key_prefix, obj->key, type_sql_map[obj->type]);
		break;
	case SIG_INSERT_FIELDS:
		ret = snprintf(buffer + offset, size - offset,
				"`%s%s`,", key_prefix, obj->key);
		break;
	case SIG_INSERT_VALUES:
		ret = snprintf(buffer + offset, size - offset, "?,");
		break;
	case SIG_WHERE_QUERY:
		op = NULL;
		if (obj->key && obj->key[0] == '$')
			op = get_query_op(obj->key);

		ret = snprintf(buffer + offset, size - offset, "%s `%.*s%s` %s ? ",
				(offset == 0 ? "WHERE" : "AND"),
				strlen(key_prefix) - (op ? 1 : 0),
				key_prefix, op ? "" : obj->key,
				op ? op : "=");
		break;
	}

	if (ret < 0 || ret >= (int)size - offset)
		ret = -ENOMEM;

out:
	free(prefix);
	return ret;
}

static inline int get_object_signature(struct faultd_object *obj, char *buffer, int size, int type)
{
	int ret;
	int off = 0;

	buffer[0] = '\0';
	if (type == SIG_CREATE_FIELDS) {
		ret = snprintf(buffer, size, "id INTEGER PRIMARY KEY,");
		if (ret < 0 || ret >= size)
			return -ENOMEM;

		off = ret;
	}

	ret = _get_object_signature(obj, buffer, size, off, type, "");
	if (ret < 0)
		return ret;

	if (off + ret > 0)
		buffer[off + ret - 1] = '\0';
	return 0;
}

static int get_table_match(struct faultd_database_adapter *adapter, struct faultd_object *obj, char *signature, char **name, int *id)
{
	sqlite3_stmt *query;
	int ret;
	char *nname;
	struct sqlite_adapter *da = to_sqlite_adapter(adapter);
	char query_str[BUFSIZE];
	char buf[BUFSIZE];
	char *sqlite_err;
	int _id;

	query = da->stmts[STMT_FIND_TABLE];
	sqlite3_reset(query);
	sqlite3_bind_text(query, 1, signature, -1, SQLITE_TRANSIENT);
	log_debug("query: %s", sqlite3_expanded_sql(query));
	ret = sqlite3_step(query);
	if (ret == SQLITE_ROW) {
		log_debug("found matching row: %s", sqlite3_column_text(query, 0));
		*name = strdup((const char *)sqlite3_column_text(query, 0));
		*id = sqlite3_column_int(query, 1);
		return 0;
	}

	log_debug("Did not found matching table. Creating new table.");

	query = da->stmts[STMT_INSERT_TABLE];
	sqlite3_reset(query);
	sqlite3_bind_text(query, 1, signature, -1, SQLITE_TRANSIENT);
	log_debug("query: %s", sqlite3_expanded_sql(query));
	ret = sqlite3_step(query);
	if (ret == SQLITE_ERROR) {
		log_error("SQL Error: %s", sqlite3_errmsg(da->db));
		return ret;
	}

	_id = sqlite3_last_insert_rowid(da->db);
	ret = asprintf(&nname, OBJ_TABLE_PREFIX "%d", _id);
	if (ret < 0)
		return -ENOMEM;

	get_object_signature(obj, buf, BUFSIZE, SIG_CREATE_FIELDS);
	ret = snprintf(query_str, BUFSIZE, "CREATE TABLE %s (%s);", nname, buf);
	if (ret < 0 || ret >= BUFSIZE)
		goto free_name;

	log_debug("creating table:\n%s", query_str);
	ret = sqlite3_exec(da->db, query_str, NULL, 0, &sqlite_err);
	if (ret != SQLITE_OK) {
		log_error("SQL Error: %s", sqlite3_errmsg(da->db));
		goto free_name;
	}

	ret = snprintf(query_str, BUFSIZE, "CREATE UNIQUE INDEX %s_oid_idx on %s (%s);", nname, nname, ID_KEY);
	if (ret < 0 || ret >= BUFSIZE)
		goto free_name;

	ret = sqlite3_exec(da->db, query_str, NULL, 0, &sqlite_err);
	if (ret != SQLITE_OK) {
		log_error("SQL Error: %s", sqlite3_errmsg(da->db));
		goto free_name;
	}

	*name = nname;
	if (id)
		*id = _id;

	return 0;

free_name:
	free(nname);
	return ret;
}

static inline uint64_t pack_oid(faultd_oid_t *oid)
{
	return (uint64_t)oid->sqlite.table << 32 | oid->sqlite.obj;
}

static inline void unpack_oid(faultd_oid_t *oid, uint64_t packed)
{
	oid->sqlite.table = (uint32_t)(packed >> 32);
	oid->sqlite.obj = (uint32_t)packed;
}

static int bind_values(struct faultd_object *obj, sqlite3_stmt *query, int n)
{
	struct faultd_object *child;
	int ret;
	int nn;
	uint64_t packed;

	switch (obj->type) {
	case TYPE_OBJECT:
		nn = n;
		list_for_each_entry(child, &obj->val.children, node) {
			ret = bind_values(child, query, nn);
			if (ret < 0)
				return ret;

			nn += ret;
		}
		ret =  nn - n;
		break;
	case TYPE_STRING:
		ret = sqlite3_bind_text(query, n, obj->val.s, -1, SQLITE_TRANSIENT);
		ret = (ret == SQLITE_OK ? 1 : -ret);
		break;
	case TYPE_OID:
		packed = pack_oid(&obj->val.oid);
		ret = sqlite3_bind_int64(query, n, packed);
		ret = (ret == SQLITE_OK ? 1 : -ret);
		break;
	case TYPE_INT:
		ret = sqlite3_bind_int(query, n, obj->val.i);
		ret = (ret == SQLITE_OK ? 1 : -ret);
		break;
	case TYPE_TIMESPEC:
		ret = sqlite3_bind_blob(query, n, (const void *)&obj->val.ts, sizeof(obj->val.ts), SQLITE_TRANSIENT);
		ret = (ret == SQLITE_OK ? 1 : -ret);
		break;
	case TYPE_TIME_T:
		ret = sqlite3_bind_int(query, n, obj->val.time);
		ret = (ret == SQLITE_OK ? 1 : -ret);
		break;
	case TYPE_UUID:
		ret = sqlite3_bind_blob(query, n, (const void *)&obj->val.uuid, sizeof(obj->val.uuid), SQLITE_TRANSIENT);
		ret = (ret == SQLITE_OK ? 1 : -ret);
		break;
	default:
		ret = -EINVAL;
		break;
	}

	return ret;
}

static int sqlite_store(struct faultd_database_adapter *adapter, struct faultd_object *obj, faultd_oid_t *oid)
{
	sqlite3_stmt *query;
	struct sqlite_adapter *da = to_sqlite_adapter(adapter);
	char fields_str[BUFSIZE];
	char values_str[BUFSIZE];
	char query_str[BUFSIZE];
	char *name = NULL;
	int ret;
	int table_id = 0;

	get_object_signature(obj, fields_str, BUFSIZE, SIG_KEY);
	ret = get_table_match(adapter, obj, fields_str, &name, &table_id);
	if (ret < 0)
		return ret;

	log_debug("storing: %s in %s", fields_str, name);
	ret = get_object_signature(obj, fields_str, BUFSIZE, SIG_INSERT_FIELDS);
	if (ret < 0) {
		log_error("Could not create query (signature: %s)", fields_str);
		goto err;
	}

	ret = get_object_signature(obj, values_str, BUFSIZE, SIG_INSERT_VALUES);
	if (ret < 0) {
		log_error("Could not create query (signature: %s)", values_str);
		goto err;
	}

	ret = snprintf(query_str, BUFSIZE, "INSERT OR REPLACE INTO %s (%s) VALUES (%s);", name, fields_str, values_str);
	if (ret < 0 || ret >= BUFSIZE)
		goto err;

	ret = sqlite3_prepare_v2(da->db, query_str, -1, &query, NULL);
	if (ret != SQLITE_OK) {
		log_error("SQL Error: %s", sqlite3_errmsg(da->db));
		goto err;
	}

	ret = bind_values(obj, query, 1);
	if (ret < 0) {
		log_error("Could not bind values to sql query");
		return ret;
	}

	log_debug("query: %s", sqlite3_expanded_sql(query));
	ret = sqlite3_step(query);
	if (ret == SQLITE_ERROR) {
		log_error("SQL Error: %s", sqlite3_errmsg(da->db));
		ret = -1;
		goto err;
	}

	if (oid) {
		oid->sqlite.table = table_id;
		oid->sqlite.obj = sqlite3_last_insert_rowid(da->db);
	}

	return 0;
err:

	free(name);
	return ret;
}

static int _build_faultd_object(sqlite3_stmt *query, int *idx, const char *signature,
		int offset, struct faultd_object *result)
{
	int p = offset, start;
	char *key = NULL;
	unsigned int i;
	int type;
	int ret;
	struct faultd_object *nobj;
	uint64_t oid_packed;
	union faultd_object_value val;

	if (signature[p++] != OBJ_SIGNATURE_BEGIN) {
		log_debug("expected begin of object signature");
		return -EINVAL;
	}

	while (signature[p] != OBJ_SIGNATURE_END) {
		start = p;
		while (signature[p] && signature[p] != ':')
			++p;

		key = strndup(&signature[start], p - start);
		++p;

		type = -1;
		for (i = 0; i < sizeof(type_id_map); ++i)
			if (signature[p] == type_id_map[i]) {
				type = i;
				break;
			}

		if (type < 0) {
			ret = -EINVAL;
			log_debug("Unknown type %c", signature[p]);
			goto err;
		}

		if (type != TYPE_OBJECT)
			++p;

		switch (type) {
		case TYPE_OBJECT:
			ret = faultd_object_new(&nobj);
			if (ret < 0)
				goto err;

			ret = _build_faultd_object(query, idx, signature, p, nobj);
			if (ret < 0) {
				faultd_object_unref(nobj);
				goto err;
			}

			p = ret;

			faultd_object_append_object(result, key, nobj);
			faultd_object_unref(nobj);
			break;
		case TYPE_STRING:
			val.s = sqlite3_column_text(query, *idx);
			faultd_object_append_string(result, key, (const char *)val.s);
			break;
		case TYPE_OID:
			oid_packed = sqlite3_column_int64(query, *idx);
			unpack_oid(&val.oid, oid_packed);
			faultd_object_append_oid(result, key, &val.oid);
			break;
		case TYPE_INT:
			val.i = sqlite3_column_int(query, *idx);
			faultd_object_append_int(result, key, val.i);
			break;
		case TYPE_TIMESPEC:
			val.ts = *(struct timespec *)sqlite3_column_blob(query, *idx);
			faultd_object_append_timespec(result, key, &val.ts);
			break;
		case TYPE_TIME_T:
			val.time = sqlite3_column_int(query, *idx);
			faultd_object_append_time_t(result, key, val.time);
			break;
		case TYPE_UUID:
			val.uuid = *(sd_id128_t *)sqlite3_column_blob(query, *idx);
			faultd_object_append_uuid(result, key, &val.uuid);
			break;
		}

		++(*idx);
		if (signature[p] == ',')
			++p;

		free(key);
		key = NULL;
	}

	return p;

err:
	free(key);
	return ret;
}

static int build_faultd_object(sqlite3_stmt *query, int *idx, const char *signature,
		int offset, struct faultd_object *result, int table_id)
{
	faultd_oid_t o;

	o.sqlite.obj = sqlite3_column_int(query, *idx);
	o.sqlite.table = table_id;
	faultd_object_append_oid(result, ID_KEY, &o);
	++(*idx);

	return _build_faultd_object(query, idx, signature, offset, result);
}

static int get_from_table(struct sqlite_adapter *da, faultd_oid_t *oid, const char *name, const char *signature,
		struct faultd_object *result)
{
	sqlite3_stmt *query;
	int ret;
	int idx = 0;
	char query_str[BUFSIZE];

	ret = snprintf(query_str, BUFSIZE,  "SELECT * FROM %s WHERE " ID_KEY "=%d;", name, oid->sqlite.obj);
	if (ret < 0 || ret >= BUFSIZE)
		return -ENOMEM;

	log_debug("%s", query_str);
	ret = sqlite3_prepare_v2(da->db, query_str, -1, &query, NULL);
	if (ret != SQLITE_OK) {
		log_error("SQL Error: %s", sqlite3_errmsg(da->db));
		return ret;
	}

	ret = sqlite3_step(query);
	if (ret != SQLITE_ROW) {
		log_debug("No entry in %s", name);
		return -ENOENT;
	}

	log_debug("found matching entry in %s", name);

	ret = build_faultd_object(query, &idx, signature, 1, result, oid->sqlite.table);
	if (ret < 0) {
		log_error_errno(ret, "could not build faultd_object: %m");
		return ret;
	}

	return 0;
}

static int sqlite_get_by_oid(struct faultd_database_adapter *adapter, faultd_oid_t *oid, struct faultd_object *result)
{
	sqlite3_stmt *query;
	struct sqlite_adapter *da = to_sqlite_adapter(adapter);
	int ret;
	char *name;
	const char *signature;

	log_debug("requesting oid %d.%d", oid->sqlite.table, oid->sqlite.obj);

	if (oid->sqlite.table != WK_TABLE_IDX) {
		ret = asprintf(&name, OBJ_TABLE_PREFIX "%d", oid->sqlite.table);
		if (ret < 0)
			return ret;

		query = da->stmts[STMT_GET_TABLE_BY_ID];
		sqlite3_reset(query);
		sqlite3_bind_int(query, 1, oid->sqlite.table);
		log_debug("query: %s", sqlite3_expanded_sql(query));
		ret = sqlite3_step(query);
		if (ret != SQLITE_ROW) {
			log_debug("table with id %d not found: %s", oid->sqlite.table, name);
			ret = -ENOENT;
			goto free_name;
		}

		signature = (const char *)sqlite3_column_text(query, 1);
		if (!signature) {
			ret = -ENOMEM;
			goto free_name;
		}
	} else {
		if (oid->sqlite.obj == WK_OID_INVALID || oid->sqlite.obj >= WK_OID_END)
			return -EINVAL;

		signature = well_known_signatures[oid->sqlite.obj];
		if (!signature)
			return -ENOMEM;

		query = da->stmts[STMT_FIND_TABLE];
		sqlite3_reset(query);
		sqlite3_bind_text(query, 1, signature, -1, SQLITE_TRANSIENT);
		log_debug("query: %s", sqlite3_expanded_sql(query));
		ret = sqlite3_step(query);
		if (ret != SQLITE_ROW) {
			log_debug("Table for well-known object does not exist yet.\n");
			return -ENOENT;
		}

		log_debug("found matching row: %s", sqlite3_column_text(query, 0));
		name = strdup((const char *)sqlite3_column_text(query, 0));
		if (!name)
			return -ENOMEM;

		oid->sqlite.table = sqlite3_column_int(query, 1);
	}

	ret = get_from_table(da, oid, name, signature, result);

free_name:
	free(name);
	return ret;
}

static int sqlite_load(struct faultd_database_adapter *adapter, struct faultd_object *query,
					 struct faultd_object *hints, struct faultd_object *result, uint32_t *nr)
{
	sqlite3_stmt *sql_query, *subquery;
	struct sqlite_adapter *da = to_sqlite_adapter(adapter);
	int ret;
	const char *name, *signature;
	char qsignature[BUFSIZE];
	char query_str[BUFSIZE];
	uint32_t n = 0;
	int table_id;

	ret = get_object_signature(query, qsignature, BUFSIZE, SIG_LIKE_PATTERN);
	if (ret < 0) {
		log_error_errno(ret, "Could not create database query: %m");
		return ret;
	}

	ret = snprintf(query_str, BUFSIZE, "SELECT ('" OBJ_TABLE_PREFIX "'||id), signature, id FROM " META_TABLE
			" WHERE signature LIKE '%s%%';", qsignature);
	if (ret < 0 || ret >= BUFSIZE)
		return -ENOMEM;

	ret = sqlite3_prepare_v2(da->db, query_str, -1, &sql_query, NULL);
	if (ret != SQLITE_OK) {
		log_error("SQL Error: %s", sqlite3_errmsg(da->db));
		return ret;
	}

	log_debug("query: %s", sqlite3_expanded_sql(sql_query));

	ret = get_object_signature(query, qsignature, BUFSIZE, SIG_WHERE_QUERY);
	if (ret < 0) {
		log_error_errno(ret, "Could not create database query: %m");
		return ret;
	}

	while ((ret = sqlite3_step(sql_query)) == SQLITE_ROW) {
		name = (const char *)sqlite3_column_text(sql_query, 0);
		signature = (const char *)sqlite3_column_text(sql_query, 1);
		table_id = sqlite3_column_int(sql_query, 2);

		ret = snprintf(query_str, BUFSIZE, "SELECT * FROM %s %s;", name, qsignature);
		if (ret < 0 || ret >= BUFSIZE)
			return ret;

		log_debug("query: %s", query_str);
		ret = sqlite3_prepare_v2(da->db, query_str, -1, &subquery, NULL);
		if (ret != SQLITE_OK) {
			log_error("SQL Error: %s", sqlite3_errmsg(da->db));
			return ret;
		}

		ret = bind_values(query, subquery, 1);
		if (ret < 0) {
			log_error_errno(ret, "could not bind values to query");
		}

		log_debug("query: %s", sqlite3_expanded_sql(subquery));

		while ((ret = sqlite3_step(subquery)) == SQLITE_ROW) {
			int idx = 0;
			struct faultd_object *part_result;
			char key[11]; /* strlen(itoa(UINT_MAX, key, 10)) => 10 */

			/* FIXME do count instead of select when result is NULL*/
			if (!result) {
				n++;
				continue;
			}

			log_debug("fetched row %u", n);
			ret = snprintf(key, sizeof(key), "%u", n);
			if (ret < 0 || ret >= sizeof(key))
				return -ENOMEM;

			ret = faultd_object_new(&part_result);
			if (ret < 0) {
				log_error_errno(ret, "could not create faultd object");
				goto finish;
			}

			ret = build_faultd_object(subquery, &idx, signature, 1, part_result, table_id);
			if (ret < 0) {
				log_error_errno(ret, "could not build faultd_object: %m");
				goto finish;
			}

			faultd_object_append_object(result, key, part_result);
			faultd_object_unref(part_result);
			n++;
		}
	}

	ret = 0;

finish:
	if (ret == 0 && nr != NULL)
		*nr = n;

	return ret;
}

static int sqlite_get_well_known_oid(const char *name, faultd_oid_t *oid)
{
	int i;
	static const struct {
		const char *name;
		int oid;
	} oids[] = {
		{FAULTD_LAST_BOOT_OID, WK_LAST_BOOT},
		{NULL, 0}
	};

	for (i = 0; oids[i].name; ++i) {
		if (strcmp(oids[i].name, name))
			continue;

		oid->sqlite.table = 0; /*FIXME: get id by signature */
		oid->sqlite.obj = oids[i].oid;
		return 0;
	}

	return -ENOENT;
}

static bool sqlite_is_oid_valid(faultd_oid_t *oid)
{
	return oid->sqlite.table > 0 && oid->sqlite.obj != WK_OID_INVALID;
}

static struct sqlite_adapter sqlite_adapter = {
	.database_adapter = {
		.module = {
			.name = "sqlite_dbadapter",
			.type = FAULTD_MODULE_TYPE_DBADAPTER,
			.init = init_sqlite_adapter,
			.cleanup = cleanup_sqlite_adapter,
			.node = LIST_HEAD_INIT(sqlite_adapter.database_adapter.module.node),
		},
		.id_key = ID_KEY,
		.store = sqlite_store,
		.get_by_oid = sqlite_get_by_oid,
		.load = sqlite_load,
		.get_well_known_oid = sqlite_get_well_known_oid,
		.is_oid_valid = sqlite_is_oid_valid,
	}
};

FAULTD_MODULE_REGISTER(&sqlite_adapter.database_adapter.module);
