/*
 * This file is part of faultd.
 *
 * Copyright © 2017 Samsung Electronics
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#include <malloc.h>
#include <errno.h>

#include "resource_violation_event.h"
#include "decision_made_event.h"
#include "action.h"
#include "event_processor.h"
#include "log.h"
#include "service.h"

#define MODULE_NAME "resource_violation_decision_maker"

static int rv_event_match(struct faultd_event_handler *handler,
						  struct faultd_event *ev)
{
	/* For now we handle here only resource violation events */
	return faultd_event_is_of_type(ev, RESOURCE_VIOLATION_EVENT_ID);
}

/*
 * This handles resource violation for all types of sevices
 * but action to be executed depends on service type:
 * - For VIP service we reboot the machine
 * - For all other services we just restart the service itself
 */
static int rv_make_decision(struct faultd_event_handler *handler)
{
	struct faultd_event *ev = pop_faultd_event(&handler->event_queue);
	struct resource_violation_event *rv_ev = to_resource_violation_event(ev);
	struct faultd_event *new_ev;
	struct dm_event_data ev_data = {
		.reason = ev,
		.who_made = MODULE_NAME,
		.action = FAULTD_ACTION_SERVICE_RESTART_ID,
	};
	int ret;

	faultd_object_new(&ev_data.action_data);
	if (systemd_service_is_of_type(&rv_ev->service, FAULTD_SERVICE_TYPE_VIP)) {
		ev_data.action = FAULTD_ACTION_REBOOT_ID;
		ret = faultd_fill_for_reboot(ev_data.action_data);
	} else {
		ret = faultd_fill_for_srv_restart(ev_data.action_data,
									  rv_ev->service.dbus_path);
	}

	if (ret) {
		log_error("Unable to create action data");
		goto cleanup;
	}

	ret = faultd_event_create(DECISION_MADE_EVENT_ID, &ev_data, &new_ev);
	if (ret) {
		log_error("Unable to create event");
		goto cleanup;
	}

	ret = event_processor_report_event(new_ev);
	faultd_event_unref(new_ev);
	if (ret) {
		log_error("Unable to report event");
		goto cleanup;
	}

cleanup:
	faultd_object_unref(ev_data.action_data);
	faultd_event_unref(ev);
	return 0;
}

static struct faultd_event_handler resource_violation_event_handler = {
	.name = MODULE_NAME,
	.event_match = rv_event_match,
	.handle_event = rv_make_decision,

	.node = LIST_HEAD_INIT(resource_violation_event_handler.node),
};

FAULTD_EVENT_HANDLER_REGISTER(resource_violation_event_handler,
							  resource_violation_eh,
							  FAULTD_MODULE_TYPE_DECISION_MAKER)
