/*
 * This file is part of faultd.
 *
 * Copyright © 2017 Samsung Electronics
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#include <malloc.h>
#include <errno.h>

#include "service_failed_event.h"
#include "decision_made_event.h"
#include "system_booted_event.h"
#include "action.h"
#include "event_processor.h"
#include "log.h"
#include "service.h"
#include "common.h"
#include "database.h"
#include "json-config.h"

#define MODULE_NAME "standard_fault_decision_maker"

struct standard_fault_decision_maker {
	struct faultd_event_handler eh;
	int n_fails_before_reboot;
};

static int sf_event_match(struct faultd_event_handler *handler,
						  struct faultd_event *ev)
{
	struct service_failed_event *sf_ev = to_service_failed_event(ev);

	return faultd_event_is_of_type(ev, SERVICE_FAILED_EVENT_ID)
		&& systemd_service_is_of_type(&sf_ev->service, FAULTD_SERVICE_TYPE_NORMAL);
}

static int sf_make_decision(struct faultd_event_handler *handler)
{
	struct faultd_event *ev = pop_faultd_event(&handler->event_queue);
	struct service_failed_event *sf_ev = to_service_failed_event(ev);
	struct standard_fault_decision_maker *sf_dm = container_of(handler, struct standard_fault_decision_maker, eh);
	struct system_booted_event *boot_event;
	struct faultd_event *new_ev;
	struct dm_event_data ev_data = {
		.reason = ev,
		.who_made = MODULE_NAME,
	};
	int ret;
	struct faultd_object *query, *service_obj;
	uint32_t nr;

	boot_event = get_boot_event();
	if (!boot_event) {
		log_error("Unable to get boot event");
		goto unref_ev;
	}

	ret = faultd_object_new(&query);
	if (ret < 0) {
		log_error("Unable to create faultd object");
		faultd_event_unref(&boot_event->event);
		goto unref_ev;
	}

	faultd_object_append_oid(query, EV_BOOT_EVENT, &boot_event->event.oid);
	faultd_event_unref(&boot_event->event);
	faultd_object_append_string(query, EV_TYPE_NAME, ev->type->name);

	ret = faultd_object_new(&service_obj);
	if (ret < 0) {
		log_error("Unable to create faultd object");
		faultd_object_unref(query);
		goto unref_ev;
	}

	faultd_object_append_string(service_obj, SYSD_SERVICE_DBUS_PATH, sf_ev->service.dbus_path);
	faultd_object_append_object(query, SYSD_SERVICE, service_obj);
	faultd_object_unref(service_obj);

	ret = database_load(query, NULL, NULL, &nr);
	faultd_object_unref(query);
	if (ret < 0) {
		log_error("Unable to load database");
		goto unref_ev;
	}

	log_debug("service failed %d times during current boot", nr);

	ret = faultd_object_new(&ev_data.action_data);
	if (ret < 0) {
		log_error("Unable to create faultd object");
		goto unref_ev;
	}

	if (nr < (uint32_t) sf_dm->n_fails_before_reboot) {
		ev_data.action = FAULTD_ACTION_SERVICE_RECOVER_ID;
		ret = faultd_fill_for_srv_recover(ev_data.action_data,
								  sf_ev->service.dbus_path,
								  sf_ev->service.recovery_unit);
	} else {
		ev_data.action = FAULTD_ACTION_REBOOT_ID;
		ret = faultd_fill_for_reboot(ev_data.action_data);
	}

	if (ret) {
		log_error("Unable to create action data");
		goto unref_action_data;
	}

	ret = faultd_event_create(DECISION_MADE_EVENT_ID, &ev_data, &new_ev);
	if (ret) {
		log_error("Unable to create event");
		goto unref_action_data;
	}

	ret = event_processor_report_event(new_ev);
	faultd_event_unref(new_ev);
	if (ret) {
		log_error("Unable to report event");
		goto unref_action_data;
	}

unref_action_data:
	faultd_object_unref(ev_data.action_data);
unref_ev:
	faultd_event_unref(ev);
	return 0;
}

static int sf_init(struct faultd_event_handler *handler, struct faultd_config *config)
{
	struct standard_fault_decision_maker *sf_dm = container_of(handler,
															   struct standard_fault_decision_maker,
															   eh);
	int ret;

	if (config == NULL)
		return 0;

	ret = get_config_field(config->root, "n_fails_before_reboot",
						   &sf_dm->n_fails_before_reboot,
						   json_type_int);
	return ret;
};

static struct standard_fault_decision_maker standard_fault_dm = {
	.eh = {
		.name = MODULE_NAME,
		.init = sf_init,
		.event_match = sf_event_match,
		.handle_event = sf_make_decision,

		.node = LIST_HEAD_INIT(standard_fault_dm.eh.node),
	},
	.n_fails_before_reboot = 3,
};

FAULTD_EVENT_HANDLER_REGISTER(standard_fault_dm.eh,
							  standard_fault_eh,
							  FAULTD_MODULE_TYPE_DECISION_MAKER)
