/*
 * This file is part of faultd.
 *
 * Copyright © 2017 Samsung Electronics
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <malloc.h>
#include <errno.h>

#include "system_booted_event.h"
#include "service_failed_event.h"
#include "decision_made_event.h"
#include "action_executed_event.h"
#include "action.h"
#include "event_processor.h"
#include "log.h"
#include "service.h"
#include "json-config.h"

#define MODULE_NAME "standard_reboot_decision_maker"

struct standard_reboot_decision_maker {
	struct faultd_event_handler eh;
	int n_reboots_before_recovery;
};

static int sr_event_match(struct faultd_event_handler *handler,
						  struct faultd_event *ev)
{
	return faultd_event_is_of_type(ev, SYSTEM_BOOTED_EVENT_ID);
}

static int get_reboot_reason(faultd_oid_t *boot_event, char **service_name)
{
	struct faultd_object *query = NULL;
	struct faultd_object *query_hints = NULL;
	struct faultd_object *orderby = NULL;
	struct faultd_object *result = NULL;
	uint32_t nr;
	int ret;

	ret = faultd_object_new_flags(&query, FAULTD_FLAG_QUERY_MODE);
	if (ret < 0)
		return ret;

	faultd_object_append_oid(query, EV_BOOT_EVENT, boot_event);
	faultd_object_append_string(query, EV_TYPE_NAME, ACTION_EXECUTED_EVENT_ID);
	faultd_object_append_string(query, AE_EV_ACTION, FAULTD_ACTION_REBOOT_ID);

	ret = faultd_object_new_flags(&query_hints, FAULTD_FLAG_QUERY_MODE);
	if (ret < 0)
		goto unref_query;

	ret = faultd_object_new(&orderby);
	if (ret < 0)
		goto unref_query_hints;

	faultd_object_append_int(orderby, EV_TIMESTAMP ".tv_sec", 1);
	faultd_object_append_object(query_hints, "$orderby", orderby);

	ret = faultd_object_new(&result);
	if (ret < 0)
		goto unref_query_hints;

	ret = database_load(query, query_hints, result, &nr);
	if (ret >= 0 && nr == 0)
		ret = -ENOENT;
	if (ret < 0)
		goto unref_result;

	faultd_object_unref(query);
	faultd_object_unref(query_hints);

	ret = faultd_object_get_string(result,
								   "0." AE_EV_ACTION_LOG "." FAULTD_AD_SERVICE_NAME,
								   service_name);
	if (ret < 0) {
		*service_name = NULL;
		faultd_object_unref(result);
		return ret;
	}

	*service_name = strdup(*service_name);
	faultd_object_unref(result);
	if (*service_name == NULL)
		return -ENOMEM;

	return 0;

unref_result:
	faultd_object_unref(result);
unref_query_hints:
	faultd_object_unref(query_hints);
unref_query:
	faultd_object_unref(query);
	return ret;
}

static int sr_make_decision(struct faultd_event_handler *handler)
{
	struct faultd_event *ev = pop_faultd_event(&handler->event_queue);
	struct standard_reboot_decision_maker *sr_dm = container_of(handler,
																struct standard_reboot_decision_maker,
																eh);
	struct faultd_event *new_ev = NULL;
	struct dm_event_data ev_data = {
		.reason = ev,
		.who_made = MODULE_NAME,
		.action = FAULTD_ACTION_RECOVERY_REBOOT_ID,
	};
	struct system_booted_event *sb_ev = NULL;
	struct faultd_event *tmp_ev = NULL;
	char *service_name = NULL;
	char *last_service_name = NULL;
	int ret;

	if (sr_dm->n_reboots_before_recovery < 0)
		goto unref_ev;

	sb_ev = get_boot_event();
	if (sb_ev == NULL) {
		log_error("Unable to get current boot event");
		goto unref_ev;
	}

	for (int i = 0; i < sr_dm->n_reboots_before_recovery; ++i) {
		/* get prev boot event */
		ret = database_get_event(&sb_ev->prev_boot_event, &tmp_ev);
		faultd_event_unref(&sb_ev->event);
		if (ret < 0) {
			log_error("Unable to get event from database");
			goto unref_ev;
		}

		sb_ev = to_system_booted_event(tmp_ev);

		/* get name of service which caused the reboot */
		ret = get_reboot_reason(&sb_ev->event.oid, &service_name);
		if (ret < 0) {
			if (ret == -ENOENT)
				log_debug("Unable to find reboot reason."
						  " Assuming user-triggered reboot.");
			else
				log_debug("Unable to get reboot reason service name");

			free(last_service_name);
			faultd_event_unref(&sb_ev->event);
			goto unref_ev;
		}

		/* compare service names */
		if (last_service_name != NULL &&
			strcmp(service_name, last_service_name) != 0) {

			free(service_name);
			free(last_service_name);
			faultd_event_unref(&sb_ev->event);
			goto unref_ev;
		}

		free(last_service_name);
		last_service_name = service_name;
	}

	free(last_service_name);
	faultd_event_unref(&sb_ev->event);

	/* make decision */
	ret = faultd_object_new(&ev_data.action_data);
	if (ret < 0) {
		log_error("Unable to create faultd object");
		goto unref_ev;
	}

	ret = faultd_fill_for_recovery_reboot(ev_data.action_data);
	if (ret < 0) {
		log_error("Unable to create action data");
		faultd_object_unref(ev_data.action_data);
		goto unref_ev;
	}

	ret = faultd_event_create(DECISION_MADE_EVENT_ID, &ev_data, &new_ev);
	faultd_object_unref(ev_data.action_data);
	if (ret < 0) {
		log_error("Unable to create event");
		goto unref_ev;
	}

	ret = event_processor_report_event(new_ev);
	faultd_event_unref(new_ev);
	if (ret) {
		log_error("Unable to report event");
		goto unref_ev;
	}

unref_ev:
	faultd_event_unref(ev);
	return 0;
}

static int sr_init(struct faultd_event_handler *handler, struct faultd_config *config)
{
	struct standard_reboot_decision_maker *sr_dm = container_of(handler,
																struct standard_reboot_decision_maker,
																eh);
	int ret;

	if (config == NULL)
		return 0;

	ret = get_config_field(config->root, "n_reboots_before_recovery",
						   &sr_dm->n_reboots_before_recovery,
						   json_type_int);
	return ret;
};

static struct standard_reboot_decision_maker standard_reboot_dm = {
	.eh = {
		.name = MODULE_NAME,
		.init = sr_init,
		.event_match = sr_event_match,
		.handle_event = sr_make_decision,

		.node = LIST_HEAD_INIT(standard_reboot_dm.eh.node),
	},
	.n_reboots_before_recovery = 3,
};

FAULTD_EVENT_HANDLER_REGISTER(standard_reboot_dm.eh,
							  standard_reboot_eh,
							  FAULTD_MODULE_TYPE_DECISION_MAKER)
