/*
 * This file is part of faultd.
 *
 * Copyright © 2017 Samsung Electronics
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#include <malloc.h>
#include <errno.h>

#include "service_failed_event.h"
#include "decision_made_event.h"
#include "action.h"
#include "event_processor.h"
#include "log.h"
#include "service.h"

#define MODULE_NAME "vip_fault_decision_maker"

static int vf_event_match(struct faultd_event_handler *handler,
						  struct faultd_event *ev)
{
	struct service_failed_event *sf_ev = to_service_failed_event(ev);

	return faultd_event_is_of_type(ev, SERVICE_FAILED_EVENT_ID)
		&& systemd_service_is_of_type(&sf_ev->service, FAULTD_SERVICE_TYPE_VIP);
}

static int vf_make_decision(struct faultd_event_handler *handler)
{
	struct faultd_event *ev = pop_faultd_event(&handler->event_queue);
	struct faultd_event *new_ev;
	struct dm_event_data ev_data = {
		.reason = ev,
		.who_made = MODULE_NAME,
		.action = FAULTD_ACTION_REBOOT_ID,
	};
	int ret;

	ret = faultd_object_new(&ev_data.action_data);
	if (ret < 0) {
		log_error("Unable to create faultd object");
		goto unref_ev;
	}

	ret = faultd_fill_for_reboot(ev_data.action_data);
	if (ret) {
		log_error("Unable to create action data");
		faultd_object_unref(ev_data.action_data);
		goto unref_ev;
	}

	ret = faultd_event_create(DECISION_MADE_EVENT_ID, &ev_data, &new_ev);
	faultd_object_unref(ev_data.action_data);
	if (ret) {
		log_error("Unable to create event");
		goto unref_ev;
	}

	ret = event_processor_report_event(new_ev);
	faultd_event_unref(new_ev);
	if (ret) {
		log_error("Unable to report event");
		goto unref_ev;
	}

unref_ev:
	faultd_event_unref(ev);
	return 0;
}

static struct faultd_event_handler vip_fault_event_handler = {
	.name = MODULE_NAME,
	.event_match = vf_event_match,
	.handle_event = vf_make_decision,

	.node = LIST_HEAD_INIT(vip_fault_event_handler.node),
};

FAULTD_EVENT_HANDLER_REGISTER(vip_fault_event_handler,
							  vip_fault_eh,
							  FAULTD_MODULE_TYPE_DECISION_MAKER)
