/*
 * This file is part of faultd.
 *
 * Copyright © 2017 Samsung Electronics
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>
#include <errno.h>
#include <malloc.h>

#include "resource_violation_event.h"

static int allocate_rv_event(struct faultd_event_type *type,
							 void *data, struct faultd_event **ev)
{
	struct resource_violation_event *rv_ev;
	struct rv_event_data *rv_ev_data = data;
	int ret;

	rv_ev = calloc(1, sizeof(*rv_ev));
	if (!rv_ev)
		return -ENOMEM;

	ret = faultd_event_init_internal(type, &rv_ev->event);
	if (ret)
		goto free_rv_ev;

	rv_ev->service = rv_ev_data->service;
	rv_ev->detection_time = rv_ev_data->detection_time;
	rv_ev->resource_type = rv_ev_data->resource_type;
	rv_ev->pid = rv_ev_data->pid;

	*ev = &rv_ev->event;
	return 0;
free_rv_ev:
	free(rv_ev);

	return ret;
}

static int deserialize_rv_event(struct faultd_event_type *type,
								struct faultd_object *data, struct faultd_event **ev)
{
	int ret = -EINVAL;
	struct rv_event_data rv_ev_data;
	struct faultd_object *obj;

	memset(&rv_ev_data, 0, sizeof(rv_ev_data));

	list_for_each_entry(obj, &data->val.children, node) {
		if ((obj->type == TYPE_OBJECT) &&
			(strcmp(SYSD_SERVICE, obj->key) == 0)) {

			ret = systemd_service_deserialize(obj, &rv_ev_data.service);
			if (ret < 0)
				goto finish;

		} else if ((obj->type == TYPE_TIME_T) &&
				   (strcmp(RV_EV_DTIME, obj->key) == 0)) {

			rv_ev_data.detection_time = obj->val.time;

		} else if ((obj->type == TYPE_INT) &&
				   (strcmp(RV_EV_RES_TYPE, obj->key) == 0)) {

			rv_ev_data.resource_type = obj->val.i;

		} else if ((obj->type == TYPE_INT) &&
				   (strcmp(RV_EV_PID, obj->key) == 0)) {

			rv_ev_data.pid = obj->val.i;

		}
	}

	ret = allocate_rv_event(type, &rv_ev_data, ev);
	if (ret < 0) {
		free(rv_ev_data.service.dbus_path);
		goto finish;
	}

	ret = faultd_event_deserialize_internal(data, type, *ev);
	if (ret < 0) {
		struct resource_violation_event *rv_ev =
			to_resource_violation_event(*ev);
		free(rv_ev_data.service.dbus_path);
		rv_ev_data.service.dbus_path = NULL;
		free(rv_ev);
		goto finish;
	}

	ret = 0;
finish:
	return ret;
}
TEST_ENTRY_POINT(deserialize_rv_event);

static void rv_event_release(struct faultd_event *ev)
{
	struct resource_violation_event *rv_ev =
		to_resource_violation_event(ev);

	systemd_service_cleanup(&rv_ev->service);
	faultd_event_cleanup_internal(&rv_ev->event);
	free(rv_ev);
}

static void rv_event_serialize(struct faultd_event *ev, struct faultd_object *out)
{
	struct resource_violation_event *rv_ev =
		to_resource_violation_event(ev);
	faultd_event_serialize_internal(ev, out);
	systemd_service_serialize(&rv_ev->service, out);
	faultd_object_append_time_t(out, RV_EV_DTIME, rv_ev->detection_time);
	faultd_object_append_int(out, RV_EV_RES_TYPE, rv_ev->resource_type);
	faultd_object_append_int(out, RV_EV_PID, rv_ev->pid);
}
TEST_ENTRY_POINT(rv_event_serialize);

static struct faultd_event_type resource_violation_event_type = {
	.name = RESOURCE_VIOLATION_EVENT_ID,
	.default_ops = {
		.release = rv_event_release,
		.serialize = rv_event_serialize,
	},
	.allocate_event = allocate_rv_event,
	.deserialize_event = deserialize_rv_event,
	.node = LIST_HEAD_INIT(resource_violation_event_type.node),
};

FAULTD_EVENT_TYPE_REGISTER(resource_violation_event_type, resource_violation_et)
