/*
 * This file is a part of faultd.
 *
 * Copyright © 2017 Samsung Electronics
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "log.h"
#include "module.h"
#include "event.h"
#include "event_processor.h"
#include "faultd_started_event.h"
#include "system_booted_event.h"
#include "database.h"

struct startup_listener {
	struct faultd_module module;
};

static int retrieve_last_boot_event(faultd_oid_t *boot_event)
{
	faultd_oid_t root_oid;
	struct faultd_object *root;
	int ret;

	faultd_object_new(&root);
	ret = database_get_well_known_oid(FAULTD_LAST_BOOT_OID, &root_oid);
	if (ret) {
		log_error("Could not get last boot id oid");
		goto out;
	}

	ret = database_get_by_oid(&root_oid, root);
	if (ret) {
		log_debug("No boot event found");
		ret = -ENOENT;
		goto out;
	}

	ret = faultd_object_get_oid(root, "boot_event", boot_event);
	if (ret) {
		log_error("Could not get last boot event");
		goto out;
	}

out:
	faultd_object_unref(root);
	return ret;
}

static int retrieve_boot_id(faultd_oid_t *boot_event, sd_id128_t *boot_id)
{
	struct faultd_object *obj;
	int ret;

	faultd_object_new(&obj);
	ret = database_get_by_oid(boot_event, obj);
	if (ret) {
		log_error("No boot event found");
		goto out;
	}

	ret = faultd_object_get_uuid(obj, SB_EV_BID, boot_id);
	if (ret) {
		log_error("Could not get boot id");
		goto out;
	}

out:
	faultd_object_unref(obj);
	return ret;
}

static int store_boot_event(faultd_oid_t *boot_event)
{
	faultd_oid_t oid;
	struct faultd_object *obj;
	int ret;

	ret = database_get_well_known_oid(FAULTD_LAST_BOOT_OID, &oid);
	if (ret < 0) {
		log_error_errno(ret, "Could not get oid of %s: %m", FAULTD_LAST_BOOT_OID);
		return ret;
	}

	faultd_object_new(&obj);
	faultd_object_append_oid(obj, database_get_id_key(), &oid);
	faultd_object_append_oid(obj, "boot_event", boot_event);

	ret = database_store(obj, &oid);

	faultd_object_unref(obj);
	return ret;
}

static int generate_event(const char *type, void *data, struct faultd_event **ev)
{
	int ret;

	ret = faultd_event_create(type, data, ev);
	if (ret) {
		log_error_errno(ret, "Unable to allocate an event");
		return ret;
	}

	ret = event_processor_report_event(*ev);
	if (ret)
		log_error_errno(ret, "Unable to report event");

	/* we are not unrefing the event here, because it is still needed in parent function */

	return ret;
}

static int startup_listener_init(struct faultd_module *module,
		struct faultd_config *cfg, sd_event *event)
{
	struct faultd_event *started_ev = NULL;
	struct faultd_event *booted_ev = NULL;
	struct sb_event_data sb_ev_data;
	sd_id128_t boot_id, last_boot_id;
	faultd_oid_t last_boot_event_oid;
	int ret = 0;

	ret = sd_id128_get_boot(&boot_id);
	if (ret)
		return ret;

	ret = retrieve_last_boot_event(&last_boot_event_oid);
	switch (ret) {
	case -ENOENT:
		/* database is empty */

		sb_ev_data.boot_id = boot_id;
		memset(&sb_ev_data.prev_boot_event, 0, sizeof(sb_ev_data.prev_boot_event));
		break;

	case 0:
		ret = retrieve_boot_id(&last_boot_event_oid, &last_boot_id);
		if (ret)
			return ret;

		if (!sd_id128_equal(boot_id, last_boot_id)) {
			/* first start in current boot */

			sb_ev_data.boot_id = boot_id;
			sb_ev_data.prev_boot_event = last_boot_event_oid;
		} else {
			/* next start in current boot */

			ret = database_get_event(&last_boot_event_oid, &booted_ev);
			if (ret) {
				log_error_errno(ret, "Unable to create system booted event from database");
				return ret;
			}

			booted_ev->oid = last_boot_event_oid;
			goto finish;
		}
		break;

	default:
		return ret;
		break;
	}

	ret = generate_event(SYSTEM_BOOTED_EVENT_ID, &sb_ev_data, &booted_ev);
	if (ret)
		return ret;

	ret = store_boot_event(&booted_ev->oid);
	if (ret) {
		log_error_errno(ret, "Could not store object in database");
		faultd_event_unref(booted_ev);
		return ret;
	}

finish:
	set_boot_event(to_system_booted_event(booted_ev));
	faultd_event_unref(booted_ev);

	ret = generate_event(FAULTD_STARTED_EVENT_ID, NULL, &started_ev);
	if (ret)
		return ret;

	faultd_event_unref(started_ev);
	return 0;
}

static void startup_listener_cleanup(struct faultd_module *module)
{
}

struct startup_listener startup_listener = {
	.module = {
		.name = "startup_listener",
		.type = FAULTD_MODULE_TYPE_LISTENER,

		.init = startup_listener_init,
		.cleanup = startup_listener_cleanup,
		.node = LIST_HEAD_INIT(startup_listener.module.node),
	},
};

FAULTD_MODULE_REGISTER(&startup_listener.module)
