#ifndef FAULTD_HELPERS_H
#define FAULTD_HELPERS_H

#include "fdtf.h"
#include "event.h"
#include "service.h"

#include <time.h>
#include <stdio.h>
#include <systemd/sd-id128.h>
#include <systemd/sd-bus.h>

#ifdef USE_EJDB
#include <ejdb/bson.h>
#endif

#include "system_booted_event.h"

#define STR_SIZE 20

#define GET_FAULTD_OBJECT_VAL(TYPE, VAR, OBJ, KEY, TYPE2) \
	TYPE VAR; \
	assert_int_equal(faultd_object_get_val(OBJ, KEY, TYPE2, &VAR), 0);

#define _S_ELEMENT(KEY, TYPE, DESIRED) \
	.key = KEY, \
	.type = TYPE, \
	.desired = DESIRED

#define S_ELEMENT(TYPE, DESIRED, KEY) { \
	_S_ELEMENT(TYPE, DESIRED, KEY) \
}

#define S_ELEMENT_EVENT(EV) \
	S_ELEMENT(EV_BOOT_EVENT, TYPE_OID, &EV.boot_event->event.oid), \
	S_ELEMENT(EV_TYPE_NAME, TYPE_STRING, &EV.type->name), \
	S_ELEMENT(EV_TIMESTAMP, TYPE_TIMESPEC, &EV.timestamp)

#define S_ELEMENT_SERVICE(srv) \
	S_ELEMENT(SYSD_SERVICE_DBUS_PATH, TYPE_STRING, \
			  &srv.dbus_path), \
	S_ELEMENT(SYSD_SERVICE_SERVICE_TYPE, TYPE_STRING, \
			  &srv.service_type), \
	S_ELEMENT(SYSD_SERVICE_RECOVERY_UNIT, TYPE_STRING, \
			  &srv.recovery_unit)

#define S_ELEMENT_OBJ(KEY, DESIRED, CHILD, CHILDREN_NUM) { \
	_S_ELEMENT(KEY, TYPE_OBJECT, DESIRED), \
	.child = CHILD, \
	.children_num = CHILDREN_NUM \
}

struct serialized_element {
	int type;
	void *desired;
	char *key;
	struct serialized_element *child;
	int children_num;
};

void test_serialization(struct serialized_element *data, int size,
						struct faultd_object *parent);
int faultd_object_count_children(struct faultd_object *obj);

void assert_test_attrs_equal(struct serialized_element *actual,
							 struct serialized_element *desired, int size);
void assert_faultd_object_equal(struct faultd_object *actual,
								struct faultd_object *desired);
void assert_data_equal(int type, void *actual, void *desired);
void assert_long_equal(const long actual, const long expected);
void assert_time_t_equal(const time_t actual, const  time_t expected);
void assert_timespec_equal(const struct timespec *actual,
						   const  struct timespec *expected);
void assert_oid_equal(const faultd_oid_t *actual, const faultd_oid_t *expected);
void assert_uuid_equal(const sd_id128_t *actual, const sd_id128_t *expected);

char generate_char();
char *generate_string(int len);
int generate_int();
time_t generate_time_t();
sd_id128_t generate_uuid();
faultd_oid_t generate_oid();
struct timespec generate_timespec();

#ifdef USE_EJDB
bson_oid_t generate_bson_oid_t();
#endif

struct serialize_test_data;

typedef void (*serialize_func_t)(struct faultd_event *, struct faultd_object *);
typedef int (*deserialize_func_t)(struct faultd_event_type *,
								  struct faultd_object *, struct faultd_event **);
typedef void (*generate_attrs_func_t)(struct serialize_test_data *);

struct serialize_test_data {
	struct faultd_event *ev;
	struct faultd_object *fd_obj;
	struct serialized_element *test_data;
	size_t size;
	serialize_func_t serialize_func;
	deserialize_func_t deserialize_func;
	generate_attrs_func_t generate_attrs_func;
};

void test_event_serialization(void **state);
void test_event_deserialization(void **state);

struct serialized_element *serialized_elems_dup(
	struct serialized_element *elems, size_t size);

void serialized_elems_free(struct serialized_element *elems, size_t size);

int setup_event_serialization(void **state,
							  struct faultd_event *(*alloc_event)(),
							  generate_attrs_func_t generate_attrs,
							  serialize_func_t serialize_func,
							  deserialize_func_t deserialize_func);

static inline void allocate_boot_event()
{
	struct system_booted_event *sb_ev;

	sb_ev = malloc(sizeof(struct system_booted_event));

	if (sb_ev) {
		sb_ev->event.oid = generate_oid();
		sb_ev->event.uref.refcnt = 1;
		set_boot_event(sb_ev);
	} else {
		printf("Unable to allocate boot event");
		fail();
	}
}

static inline void free_boot_event()
{
	struct system_booted_event *sb_ev = get_boot_event();
	free(sb_ev);
}

static inline void fill_event(struct faultd_event *ev)
{
	allocate_boot_event();
	ev->boot_event = get_boot_event();
	ev->type = calloc(1, sizeof(*(ev->type)));
	assert_non_null(ev->type);
	ev->type->name = generate_string(STR_SIZE);
	ev->timestamp = generate_timespec();
}

static inline void free_event(struct faultd_event *ev)
{
	free_boot_event();
	free(ev->type->name);
	free(ev->type);
}

static inline void fill_service(struct systemd_service *srv)
{
	srv->dbus_path = generate_string(STR_SIZE);
	srv->service_type = generate_string(STR_SIZE);
	srv->recovery_unit = generate_string(STR_SIZE);
}

static inline void free_service(struct systemd_service *srv)
{
	free(srv->dbus_path);
	free(srv->service_type);
	free(srv->recovery_unit);
}

static inline void free_test_data(struct serialize_test_data *td)
{
	faultd_object_unref(td->fd_obj);
	serialized_elems_free(td->test_data, td->size);
	free(td);
}

#endif /* FAULTD_HELPERS_H */
